/*------------------------------------------------------------------------------
 bin/djv_convert/app.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "app.h"
#include "etc.h"
#include <djv_image/io.h>
#include <djv_base/user.h>
#include <djv_base/sys.h>

namespace djv_convert {

App::App(int argc, char ** argv) :
  djv_image::app::App("djv_convert", argc, argv)
{
  DJV_DBG_OBJ("App::App");
  
  // Command line.

  try
  {
    Str tmp = cmdln_arg();
    cmdln(tmp);
  }
  catch (Error error)
  {
    this->error(app::error_cmdln, error);
    exit(app::EXIT_ERROR);
  }
  if (exit() != app::EXIT_DEFAULT) return;
  
  // Work.

  if (! work()) exit(app::EXIT_ERROR);
}

bool App::work()
{
  DJV_DBG_OBJ("App::work");
  DJV_DBG_PRN("input = " << _in);
  DJV_DBG_PRN("output = " << _out);

  djv_base::time::Timer timer;
  timer.start();

  djv_image::gl_data::Option option;
  option.xform.mirror = _image.mirror;
  option.xform.scale = _image.scale;
  option.channel = _image.channel;

  // Open input.

  std::auto_ptr<djv_image::io::Load> load;
  djv_image::Info load_info;
  djv_image::Tag load_tag;
  djv_image::Time load_time;

  Error error;
  while (! load.get())
  {
    try
    {
      load = std::auto_ptr<djv_image::io::Load>(
        djv_image::io::global()->load()->get(_in, &load_info, &load_tag,
        &load_time));
    }
    catch (Error in)
    {
      error = in;
    }

    if (! load.get() && _input.timeout)
    {
      print(label_timeout);
      djv_base::time::sleep(1);
    }
    else
    {
      break;
    }
  }
  if (! load.get())
  {
    this->error(error, _in.name());
    return false;
  }

  int start = 0, end = load_time.list.size() ?
    (int(load_time.list.size()) - 1) : 0;
  if (! _input.start.empty())
    start = file::seq_find(
      djv_image::time::str_to_units(_input.start, load_time.speed),
      load_time.list);
  if (! _input.end.empty())
    end = file::seq_find(
      djv_image::time::str_to_units(_input.end, load_time.speed),
      load_time.list);
  load_time.list = List<int>(load_time.list, start, end - start + 1);

  print(str::Format("%% %%").arg(_in).arg(label_image(load_info, load_time)));
  
  // Open output.
  
  std::auto_ptr<djv_image::io::Save> save;
  djv_image::Info save_info = load_info;
  djv_image::Time save_time = load_time;
  
  if (vec::size_valid(_image.size))
    save_info.size = _image.size;
  else if (_image.size.x)
    save_info.size = Vec2i(
      _image.size.x,
      math::ceil(_image.size.x / vec::aspect(load_info.size)));
  else if (_image.size.y)
    save_info.size = Vec2i(
      math::ceil(_image.size.x * vec::aspect(load_info.size)),
      _image.size.y);
  else
    save_info.size = vec::ceil<double, int>(Vec2f(load_info.size) *
      option.xform.scale);
  
  if (_output.pixel.get())
    save_info.pixel = *_output.pixel;

  save_time = djv_image::time::Time(
    _out.seq.start(),
    _out.seq.start() +
      (load_time.list.size() ? (int(load_time.list.size()) - 1) : 0));
  if (_output.speed.get())
    save_time.speed =  *_output.speed;
  
  try
  {
    save = std::auto_ptr<djv_image::io::Save>(
      djv_image::io::global()->save()->get(_out, save_info, _output.tag,
      save_time));
  }
  catch (Error error)
  {
    this->error(error, _out.name());
    return false;
  }

  print(str::Format("%% %%").arg(_out).arg(label_image(save_info, save_time)));

  // Work.
  
  _buffer = std::auto_ptr<djv_image::gl_offscreen::Buffer>(
    djv_image::gl_offscreen::global()->buffer_create());
  _buffer->init(save_info);
  
  static const int progress_length = 10;
  int progress_count = 0;
  double progress_accum = 0.0;
  
  const uint length = save_time.list.size();
  for (uint i = 0; i < length; ++i)
  {
    djv_base::time::Timer frame_timer;
    frame_timer.start();
    
    // Load.

    const djv_image::Image * image = 0;
    
    int timeout = _input.timeout;
    while (! image)
    {
      try
      {
        image = load->get(
          load_time.list.size() ? load_time.list[i] : -1,
          _input.layer,
          _input.proxy);
      }
      catch (Error in)
      {
        error = in;
      }
      
      if (! image && timeout > 0)
      {
        //print(label_timeout);
        --timeout;
        djv_base::time::sleep(1);
      }
      else
      {
        break;
      }
    }
    if (! image)
    {
      this->error(error, _in.name());
      return false;
    }
    DJV_DBG_PRN("image = " << *image);

    // Image tags.
    
    djv_image::Tag tag = _output.tag;

    tag.add(image->tag);
    
    if (_output.tag_auto)
    {
      tag[djv_image::tag::label_tag()[djv_image::tag::CREATOR]] =
        user::current();
      tag[djv_image::tag::label_tag()[djv_image::tag::TIME]] =
        djv_base::time::time_to_str(djv_base::time::current());
      tag[djv_image::tag::label_tag()[djv_image::tag::TIMECODE]] =
        djv_image::time::timecode_to_str(djv_image::time::frame_to_timecode(
          save_time.list.size() ? save_time.list[i] : 0, save_time.speed));
    }
    
    // Process.

    djv_image::Image tmp(save_info);
    tmp.tag = tag;
    
    option.xform.scale = Vec2f(save_info.size) / Vec2f(load_info.size);
    option.profile = image->profile;
    djv_image::gl_data::copy(*image, &tmp, option, &_buffer_data,
      _buffer.get());
    
    // Save.

    DJV_DBG_PRN("output = " << tmp);
    
    try
    {
      save->set(tmp, save_time.list.size() ? save_time.list[i] : -1);
    }
    catch (Error error)
    {
      this->error(error, _out.name());
      return false;
    }

    // Statistics.

    timer.check();
    frame_timer.check();
    
    progress_accum += frame_timer.seconds();
    if (length > 1 && i / double(length) * progress_length > progress_count)
    {
      const double estimate =
        progress_accum / double(i + 1) * (length - (i + 1));

      print(str::Format(label_estimate).
        arg(int(i / double(length) * 100.0), 3).
        arg(djv_base::time::label_time(estimate)).
        arg(i / timer.seconds(), 0, 2));

      ++progress_count;
    }
  }
  if (length > 1) print(label_complete, false);
  
  try
  {
    save->close();
  }
  catch (Error error)
  {  
    this->error(error, _out.name());
    return false;
  }

  print(str::Format(label_elapsed).
    arg(djv_base::time::label_time(timer.seconds())));

  return true;
}

}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = djv_convert::App(argc, argv).exec(); }
  catch (djv_base::Error in) { djv_base::error::print(in); }
  return r;
}

