/*------------------------------------------------------------------------------
 bin/djv_ls/app.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "app.h"
#include "etc.h"
#include <djv_base/dir.h>
#include <djv_base/math.h>
#include <djv_base/sys.h>
#include <djv_base/user.h>

namespace djv_ls {

App::App(int argc, char ** argv) :
  app::App("djv_ls", argc, argv),
  _seq(file::SEQ_RANGE),
  _recurse(false),
  _hidden(false),
  _columns(sys::terminal_width()),
  _sort(file::SORT_NAME),
  _sort_reverse(false),
  _sort_dir(true)
{
  DJV_DBG_OBJ("App::App");
  
  // Command line.
  
  try
  {
    Str tmp = cmdln_arg();
    cmdln(tmp);
  }
  catch (Error in)
  {
    error(app::error_cmdln, in);
    exit(app::EXIT_ERROR);
  }
  if (exit() != app::EXIT_DEFAULT) return;

  // Setup.

  List<File> list;
  for (uint i = 0; i < _input.size(); ++i)
  {
    File file(_input[i], false);
    if (_seq)
    {
      if (file.seq_wildcard())
        file = file::seq_match(file, Dir(file.path).list());
      if (file.seq_valid())
        file.type = file::SEQ;
    }

    List<Str> tmp = file::seq_expand(file);
    for (uint j = 0; j < tmp.size(); ++j)
    {
      file = File(tmp[j], false);
      if (! file.stat())
      {
        this->error(error_open, file);
        exit(app::EXIT_ERROR);
        continue;
      }
      if (_seq && file.seq_valid()) file.type = file::SEQ;
      list += file;
    }
  }
  DJV_DBG_PRN("list = " << list);

  file::seq_compress(&list, _seq);
  if (! _hidden) file::filter(&list, file::FILTER_HIDDEN, _filter);
  file::sort(&list, _sort, _sort_reverse);
  if (_sort_dir) list = file::sort_dir(list);
  DJV_DBG_PRN("list = " << list);

  if (! list.size() && exit() != app::EXIT_ERROR)
    list += File(".");

  // Work.

  for (uint i = 0; i < list.size(); ++i)
    if (file::DIR == list[i].type)
    {
      if (! dir(list[i], (list.size() > 1) || _recurse))
      {
        error(error_open, list[i]);
        exit(app::EXIT_ERROR);
      }
    }
    else
    {
      file(list[i], true);
    }
}

App::~App()
{
  DJV_DBG_OBJ("App::~App");
}

void App::file(const File & in, bool path)
{
  DJV_DBG_OBJ("App::file");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("path = " << path);
  DJV_DBG_PRN("columns = " << _columns);
  
  Str name = in.get(-1, path);
  
  const Str info_str = str::Format(label_info_format).
    arg(file::label_size(in.size)).
#if ! defined(WIN32)
    arg(user::uid_to_str(in.user)).
#endif
    arg(file::label_type()[in.type], 4).
    arg(file::label_perm(in.perm), 3).
    arg(djv_base::time::time_to_str(in.time));

  if (_columns)
  {
    const uint size = math::max(_columns - int(info_str.size()) - 2, 0);
    if (name.size() > size)
    {
      name = Str(name, 0, size);
      size_t i = name.size(), j = 3;
      while (i-- > 0 && j-- > 0) name[i] = '.';
    }
  }
  
  print(str::Format("%% %%").arg(name).
    arg(info_str, _columns - int(name.size()) - 2)
  );
}

bool App::dir(const File & in, bool label)
{
  DJV_DBG_OBJ("App::dir");
  DJV_DBG_PRN("in = " << in);
  
  // Setup.

  Dir dir(in);
  if (! dir.valid()) return false;
  
  List<File> list = dir.list();
  file::seq_compress(&list, _seq);
  if (! _hidden) file::filter(&list, file::FILTER_HIDDEN, _filter);
  file::sort(&list, _sort, _sort_reverse);
  if (_sort_dir) list = file::sort_dir(list);

  // Print.

  if (label) print(str::Format(label_dir).arg(in));

  for (uint i = 0; i < list.size(); ++i)
    file(list[i]);

  separator();
  
  // Recurse.

  bool r = true;
  if (_recurse)
  {
    List<File> list = Dir(in).list();
    file::filter(&list, file::FILTER_FILE |
      (! _hidden ? file::FILTER_HIDDEN : 0));
    for (uint i = 0; i < list.size(); ++i)
      r &= this->dir(list[i], label);
  }
  return r;
}

}

int main(int argc, char ** argv)
{
  int r = 1;
  try { r = djv_ls::App(argc, argv).exec(); }
  catch (djv_base::Error in) { djv_base::error::print(in); }
  return r;
}

