/*------------------------------------------------------------------------------
 bin/djv_view/file_group.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include "app.h"
#include "file.h"
#include "image.h"
#include "shortcut.h"
#include <djv_ui/app_dialog.h>
#include <djv_ui/button.h>
#include <djv_ui/file_browser.h>
#include <djv_ui/image_io.h>
#include <djv_ui/label.h>
#include <djv_ui/messages_dialog.h>
#include <djv_ui/prefs.h>
#include <djv_ui/prefs_dialog.h>
#include <djv_image/io.h>
#include <djv_base/dir.h>
#include <FL/Fl.H>

namespace djv_view {
namespace file {

struct Group::_Group
{
  _Group() :
    menu(0)
  {}

  djv_ui::menu::Widget * menu;
  List<int> menu_recent, menu_layer, menu_proxy;
  djv_ui::button::Tool * open, * reload, * close;
};

Group::Group(
  djv_ui::menu::Widget * menu,
  djv_ui::layout::Horizontal * toolbar,
  const Group * copy
) :
  open_sig(this),
  reload_sig(this),
  save_sig(this),
  save_frame_sig(this),
  image_sig(this),
  update_sig(this),
  _file(copy ? copy->_file : File()),
  _image(0),
  _layer(copy ? copy->_layer : 0),
  _proxy(copy ? copy->_proxy : prefs()->proxy()),
  _cache(copy ? copy->_cache : prefs()->cache()),
  _cache_ref(0),
  _p(new _Group)
{
  using namespace djv_ui;
  
  // Widgets.
  
  _p->menu = menu;
  
  _p->open = new button::Tool("file_open");
  _p->reload = new button::Tool("file_reload");
  _p->close = new button::Tool("file_close");

  // Layout.
  
  toolbar->add(_p->open);
  toolbar->add(_p->reload);
  toolbar->add(_p->close);
  
  // Initialize.
  
  if (copy) open(_file);
  
  // Callbacks.
  
  _p->open->sig.set(this, open_callback);
  _p->reload->sig.set(this, reload_callback);
  _p->close->sig.set(this, close_callback);

  prefs()->recent_sig.set(this, recent_callback);
  prefs()->proxy_sig.set(this, proxy_callback);
  prefs()->cache_sig.set(this, cache_callback);

  djv_ui::image_io::global()->sig.set(this, reload_callback);
}

Group::~Group()
{}

void Group::del()
{
  Callback_Base::del();
  
  if (_cache_ref) _cache_ref->ref_del();
  cache_update();
}

void Group::menu_update(djv_ui::menu::Group * in)
{
  using namespace djv_ui::menu;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  in->add(menu_title, 0, 0, 0, SUB_MENU);
  
  in->add(
    menu_open, shortcut[shortcut::FILE_OPEN].value,
    _open_callback, this);
  
  in->add(menu_recent, 0, 0, 0, SUB_MENU);
  _p->menu_recent = in->add(
    prefs()->recent(), List<int>(),
    _recent_callback, this);
  in->end();
  
  in->add(
    menu_reload, shortcut[shortcut::FILE_RELOAD].value,
    _reload_callback, this);
  
  in->add(
    menu_close, shortcut[shortcut::FILE_CLOSE].value,
    _close_callback, this,
    DIVIDER);

  in->add(
    menu_save, shortcut[shortcut::FILE_SAVE].value,
    _save_callback, this);
  
  in->add(
    menu_save_frame, shortcut[shortcut::FILE_SAVE_FRAME].value,
    _save_frame_callback, this,
    DIVIDER);
  
  in->add(menu_layer, 0, 0, 0, SUB_MENU);
  _p->menu_layer = in->add(
    _layer_list, List<int>() <<
      shortcut[shortcut::FILE_LAYER_DEFAULT].value <<
      shortcut[shortcut::FILE_LAYER_1].value <<
      shortcut[shortcut::FILE_LAYER_2].value <<
      shortcut[shortcut::FILE_LAYER_3].value <<
      shortcut[shortcut::FILE_LAYER_4].value <<
      shortcut[shortcut::FILE_LAYER_5].value <<
      shortcut[shortcut::FILE_LAYER_6].value <<
      shortcut[shortcut::FILE_LAYER_7].value <<
      shortcut[shortcut::FILE_LAYER_8].value <<
      shortcut[shortcut::FILE_LAYER_9].value <<
      shortcut[shortcut::FILE_LAYER_10].value,
    _layer_callback, this,
    RADIO, _layer);
  in->end();
  
  in->add(
    menu_layer_prev, shortcut[shortcut::FILE_LAYER_PREV].value,
    _layer_prev_callback, this);
  
  in->add(
    menu_layer_next, shortcut[shortcut::FILE_LAYER_NEXT].value,
    _layer_next_callback, this);

  in->add(menu_proxy, 0, 0, 0, SUB_MENU);
  _p->menu_proxy = in->add(
    djv_image::data::label_proxy(), List<int>() <<
      shortcut[shortcut::FILE_PROXY_NONE].value <<
      shortcut[shortcut::FILE_PROXY_1_2].value <<
      shortcut[shortcut::FILE_PROXY_1_4].value <<
      shortcut[shortcut::FILE_PROXY_1_8].value,
    _proxy_callback, this,
    RADIO, _proxy);
  in->end();

  in->add(
    menu_cache, 0,
    _cache_callback, this,
    TOGGLE | DIVIDER, _cache);

  in->add(djv_ui::menu::menu_messages, 0, _messages_callback, this);
  in->add(djv_ui::menu::menu_prefs, 0, _prefs_callback, this);
  in->add(djv_ui::menu::menu_info, 0, _info_callback, this);
  in->add(djv_ui::menu::menu_about, 0, _about_callback, this, DIVIDER);

  in->add(
    djv_ui::menu::menu_exit, shortcut[shortcut::EXIT].value,
    _exit_callback, this);
  
  in->end();
}

void Group::toolbar_update()
{
  using djv_ui::shortcut::serialize;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  _p->open->tooltip(str::Format(tooltip_open).
    arg(serialize(shortcut[shortcut::FILE_OPEN].value)));
  _p->reload->tooltip(str::Format(tooltip_reload).
    arg(serialize(shortcut[shortcut::FILE_RELOAD].value)));
  _p->close->tooltip(str::Format(tooltip_close).
    arg(serialize(shortcut[shortcut::FILE_CLOSE].value)));
}

void Group::open(const File & in)
{
  DJV_DBG_OBJ("file::Group::open");
  DJV_DBG_PRN("in = " << in);

  File file = in;
  if (file.seq_wildcard())
    file = djv_base::file::seq_match(file, Dir(file.path).list());
  DJV_DBG_PRN("in = " << file);

  _file = File();
  _info.clear();
  _tag = djv_image::Tag();
  _time = djv_image::Time();
  _load.reset();
  
  // Automatic sequence loading.
  
  if (prefs()->seq_auto() && file.seq_valid() && file.seq.list.size() == 1)
  {
    file.type = djv_base::file::SEQ;
    
    file.seq.list.clear();
    const List<File> list = Dir(file.path).list();
    for (uint i = 0; i < list.size(); ++i)
      file.seq_add(list[i]);

    djv_base::file::seq_sort(&file.seq);

    DJV_DBG_PRN("sequence = " << file);
  }

  // Load.
  
  if (! file.get().empty())
  {
    DJV_DBG_PRN("loading...");
    
    try
    {
      _load = std::auto_ptr<djv_image::io::Load>(
        djv_image::io::global()->load(file, &_info, &_tag, &_time));
      _file = file;

      prefs()->recent(_file);
    }
    catch (Error error)
    {
      DJV_DBG_PRN("error = " << error);
      
      app()->error(error, file.name());
    }
  }

  _layer_list.resize(_info.size());
  for (uint i = 0; i < _info.size(); ++i)
    _layer_list[i] = _info[i].name;
  
  cache_update();
}

void Group::open_callback(const File & in) { open_sig.emit(in); }

void Group::open_callback(bool)
{
  djv_ui::file_browser::set(this, open_callback);
}

void Group::_open_callback() { open_callback(true); }

const djv_image::Image * Group::get(int frame)
{
  DJV_DBG_OBJ("file::Group::get");
  DJV_DBG_PRN("frame = " << frame);
  
  _image = 0;
  image_cache::Ref * prev = _cache_ref;
  
  _cache_ref = _cache ? image_cache::global()->get(this, frame) : 0;
  if (_cache_ref)
  {
    DJV_DBG_PRN("cached image");
    
    _image = _cache_ref->get();
  }
  else
  {
    if (_load.get())
    {
      DJV_DBG_PRN("loading image");

      try
      {
        _image = _load->get(
          _time.list.size() ? _time.list[frame] : -1,
          _layer,
          _proxy);
      }
      catch (Error error)
      {
        DJV_DBG_PRN("error" << error);
        
        app()->error(error, _file.name());
      }
    }
    
    if (_image && _cache)
      _cache_ref = image_cache::global()->create(
        new djv_image::Image(*_image), this, frame);
  }
  
  if (prev) prev->ref_del();
  return _image;
}

const File & Group::file() const { return _file; }
const List<djv_image::Info> & Group::info() const { return _info; }
const djv_image::Tag & Group::tag() const { return _tag; }
const djv_image::Time & Group::time() const { return _time; }

void Group::recent_callback(bool)
{
  DJV_DBG_OBJ("file::Group::recent_callback");
  
  File file = prefs()->recent()[list::find(_p->menu->item(), _p->menu_recent)];
  if (file.seq_valid()) file.type = djv_base::file::SEQ;
  open_sig.emit(file);
}

void Group::_recent_callback() { recent_callback(true); }

void Group::recent_callback(const List<Str> &) { update_sig.emit(true); }

void Group::reload_callback(bool)
{
  DJV_DBG_OBJ("file::Group::reload_callback");
  
  reload_sig.emit(true);

  _load.reset();
  
  if (_file.get().size())
  {
    try
    {
      _load = std::auto_ptr<djv_image::io::Load>(
        djv_image::io::global()->load(_file, &_info, &_tag, &_time));
    }
    catch (Error error)
    {
      app()->error(error, _file.name());
    }
  }

  cache_update();
  image_sig.emit(true);
  update_sig.emit(true);
}

void Group::reload_callback(djv_image::io::Plugin * in)
{
  if (! _load.get()) return;
  
  DJV_DBG_OBJ("file::Group::reload_callback");
  
  if (! in)
  {
    reload_callback(true);
  }
  else if (in->name() == _load->name())
  {
    const List<Str> & option_list = in->option_list();
    for (uint i = 0; i < option_list.size(); ++i)
    {
      Str tmp = in->option(option_list[i]);
      _load->option(option_list[i], &tmp);
    }
  
    cache_update();
    image_sig.emit(true);
    update_sig.emit(true);
  }
}

void Group::_reload_callback() { reload_callback(true); }

void Group::close_callback(bool)
{
  DJV_DBG_OBJ("file::Group::close_callback");
  
  open_sig.emit(File());
}

void Group::_close_callback() { close_callback(true); }

void Group::save_callback(const File & in)
{
  DJV_DBG_OBJ("file::Group::save_callback");
  DJV_DBG_PRN("in = " << in);
  
  _file_save = in;
  
  if (djv_base::file::exists(_file_save))
  {
    DJV_DBG_PRN("exists");
    
    djv_ui::dialog::global()->question(
      dialog_save_check,
      this,
      save_check_callback);
  }
  else
  {
    save_sig.emit(_file_save);
  }
}

void Group::save_check_callback(bool)
{
  save_sig.emit(_file_save);
}

void Group::_save_callback()
{
  if (! _load.get()) return;
  djv_ui::file_browser::set(this, save_callback);
}

void Group::save_frame_callback(const File & in)
{
  _file_save = in;

  if (djv_base::file::exists(_file_save))
    djv_ui::dialog::global()->question(
      dialog_save_check,
      this,
      save_frame_check_callback);
  else
    save_frame_sig.emit(_file_save);
}

void Group::save_frame_check_callback(bool)
{
  save_frame_sig.emit(_file_save);
}

void Group::_save_frame_callback()
{
  if (! _load.get()) return;
  djv_ui::file_browser::set(this, save_frame_callback);
}

void Group::layer(int in)
{
  if (in == _layer) return;
  
  DJV_DBG_OBJ("file::Group::layer");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("layer list = " << _layer_list.size());

  _layer = math::wrap(in, 0, int(_layer_list.size()) - 1);
  DJV_DBG_PRN("layer = " << _layer);
  
  cache_update();
  image_sig.emit(true);
  update_sig.emit(true);
}

int Group::layer() const { return _layer; }

void Group::layer_callback(int in) { layer(in); }

void Group::_layer_callback()
{
  layer(list::find(_p->menu->item(), _p->menu_layer));
}

void Group::_layer_prev_callback() { layer(_layer - 1); }
void Group::_layer_next_callback() { layer(_layer + 1); }

void Group::proxy(djv_image::data::PROXY in)
{
  if (in == _proxy) return;
  
  DJV_DBG_OBJ("file::Group::proxy");
  DJV_DBG_PRN("in = " << in);

  _proxy = in;
  
  cache_update();
  image_sig.emit(true);
  update_sig.emit(true);
}

djv_image::data::PROXY Group::proxy() const { return _proxy; }

void Group::proxy_callback(djv_image::data::PROXY in) { proxy(in); }

void Group::_proxy_callback()
{
  proxy(djv_image::data::PROXY(list::find(_p->menu->item(), _p->menu_proxy)));
}

void Group::cache(bool in)
{
  if (in == _cache) return;
  
  DJV_DBG_OBJ("file::Group::cache");
  DJV_DBG_PRN("in = " << in);

  _cache = in;

  cache_update();
  image_sig.emit(true);
  update_sig.emit(true);
}

bool Group::cache() const { return _cache; }

List<int> Group::cache_list() const
{
  return image_cache::global()->frame_list(this);
}

void Group::cache_callback(bool in) { cache(in); }
void Group::_cache_callback() { cache(_p->menu->value()); }

void Group::cache_update()
{
  DJV_DBG_OBJ("file::Group::cache_update");
  
  callbacks(false);
  
  image_cache::global()->del(this);
  
  callbacks(true);
}

void Group::_messages_callback()
{
  djv_ui::messages_dialog::global()->show();
}

void Group::_prefs_callback() { djv_ui::prefs_dialog::global()->show(); }
void Group::_info_callback() { djv_ui::app_dialog::global()->info()->show(); }
void Group::_about_callback() { djv_ui::app_dialog::global()->about()->show(); }
void Group::_exit_callback() { app()->exit(); }

}
}

