/*------------------------------------------------------------------------------
 bin/djv_view/view_group.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include "shortcut.h"
#include "view.h"
#include <djv_ui/button.h>
#include <djv_ui/label.h>
#include <djv_ui/num.h>
#include <djv_ui/shortcut.h>
#include <djv_ui/style.h>

namespace djv_view {
namespace view {

struct Group::_Group
{
  _Group() :
    menu(0)
  {}

  djv_ui::menu::Widget * menu;
  List<int> menu_grid;
  djv_ui::button::Tool * zoom_in, * zoom_out, * zoom_reset, * fit;
  djv_ui::label::Widget * zoom_label;
};

Group::Group(
  djv_ui::menu::Widget * menu,
  djv_ui::layout::Horizontal * toolbar,
  const Group * copy
) :
  move_sig(this),
  center_sig(this),
  zoom_sig(this),
  zoom_reset_sig(this),
  reset_sig(this),
  fit_sig(this),
  overlay_sig(this),
  update_sig(this),
  _grid(copy ? copy->_grid : prefs()->grid()),
  _hud(copy ? copy->_hud : prefs()->hud()),
  _p(new _Group)
{
  using namespace djv_ui;
  
  // Widgets.
  
  _p->menu = menu;
  
  _p->zoom_in = new button::Tool("view_zoom_in");
  _p->zoom_in->type(button::Tool::REPEAT);
  _p->zoom_out = new button::Tool("view_zoom_out");
  _p->zoom_out->type(button::Tool::REPEAT);
  _p->zoom_reset = new button::Tool("view_zoom_reset");
  _p->fit = new button::Tool("view_fit");
  
  _p->zoom_label = new label::Widget;
  _p->zoom_label->size_str(toolbar_zoom_format);
  _p->zoom_label->align(layout::RIGHT, layout::CENTER);
  _p->zoom_label->tooltip(tooltip_zoom);
  num::Inc * zoom_inc_widget = new num::Inc;

  // Layout.
  
  toolbar->add(_p->zoom_in);
  toolbar->add(_p->zoom_out);
  toolbar->add(_p->zoom_reset);
  toolbar->add(_p->fit);
  toolbar->add_spacer(style::global()->spacing());
  toolbar->add(_p->zoom_label);
  toolbar->add_spacer(style::global()->spacing());
  toolbar->add(zoom_inc_widget);
  
  // Callbacks.
  
  _p->zoom_in->sig.set(this, zoom_in_callback);
  _p->zoom_out->sig.set(this, zoom_out_callback);
  zoom_inc_widget->inc_sig.set(this, zoom_inc_callback);
  zoom_inc_widget->dec_sig.set(this, zoom_dec_callback);
  _p->zoom_reset->sig.set(this, zoom_reset_callback);
  _p->fit->sig.set(this, fit_callback);
  
  prefs()->grid_sig.set(this, grid_callback);
  prefs()->hud_sig.set(this, hud_callback);
}

Group::~Group()
{}

void Group::zoom(double in)
{
  _p->zoom_label->set(str::Format(toolbar_zoom).arg(int(in * 100)));
}

void Group::menu_update(djv_ui::menu::Group * in)
{
  using namespace djv_ui::menu;
  
  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
    
  in->add(menu_title, 0, 0, 0, SUB_MENU);
  
  in->add(
    menu_left, shortcut[shortcut::VIEW_LEFT].value,
    _left_callback, this);
  
  in->add(
    menu_right, shortcut[shortcut::VIEW_RIGHT].value,
    _right_callback, this);
  
  in->add(
    menu_up, shortcut[shortcut::VIEW_UP].value,
    _up_callback, this);
  
  in->add(
    menu_down, shortcut[shortcut::VIEW_DOWN].value,
    _down_callback, this);
  
  in->add(
    menu_center, shortcut[shortcut::VIEW_CENTER].value,
    _center_callback, this,
    DIVIDER);
  
  in->add(
    menu_zoom_in, shortcut[shortcut::VIEW_ZOOM_IN].value,
    _zoom_in_callback, this);
  
  in->add(
    menu_zoom_out, shortcut[shortcut::VIEW_ZOOM_OUT].value,
    _zoom_out_callback, this);
  
  in->add(
    menu_zoom_reset, shortcut[shortcut::VIEW_ZOOM_RESET].value,
    _zoom_reset_callback, this,
    DIVIDER);
  
  in->add(
    menu_reset, shortcut[shortcut::VIEW_RESET].value,
    _reset_callback, this);
  
  in->add(
    menu_fit, shortcut[shortcut::VIEW_FIT].value,
    _fit_callback, this,
    DIVIDER);
  
  in->add(menu_grid, 0, 0, 0, SUB_MENU);
  _p->menu_grid = in->add(
    label_grid(), List<int>(),
    _grid_callback, this,
    RADIO, _grid);
  in->end();
  
  in->add(
    menu_hud, shortcut[shortcut::VIEW_HUD].value,
    _hud_callback, this,
    TOGGLE, _hud);
    
  in->end();
}

void Group::toolbar_update()
{
  using djv_ui::shortcut::serialize;

  const List<djv_ui::shortcut::Shortcut> & shortcut =
    shortcut::prefs()->shortcut();
  
  _p->zoom_in->tooltip(str::Format(tooltip_zoom_in).
    arg(serialize(shortcut[shortcut::VIEW_ZOOM_IN].value)));
  _p->zoom_out->tooltip(str::Format(tooltip_zoom_out).
    arg(serialize(shortcut[shortcut::VIEW_ZOOM_OUT].value)));
  _p->zoom_reset->tooltip(str::Format(tooltip_zoom_reset).
    arg(serialize(shortcut[shortcut::VIEW_ZOOM_RESET].value)));
  _p->fit->tooltip(str::Format(tooltip_fit).
    arg(serialize(shortcut[shortcut::VIEW_FIT].value)));
}

void Group::_left_callback() { move_sig.emit(Vec2i(10, 0)); }
void Group::_right_callback() { move_sig.emit(Vec2i(-10, 0)); }
void Group::_up_callback() { move_sig.emit(Vec2i(0, 10)); }
void Group::_down_callback() { move_sig.emit(Vec2i(0, -10)); }
void Group::_center_callback() { center_sig.emit(true); }

void Group::zoom_in_callback(bool) { zoom_sig.emit(2.0); }
void Group::zoom_out_callback(bool) { zoom_sig.emit(0.5); }
void Group::zoom_inc_callback(bool) { zoom_sig.emit(1.1); }
void Group::zoom_dec_callback(bool) { zoom_sig.emit(0.9); }
void Group::zoom_reset_callback(bool) { zoom_reset_sig.emit(true); }
void Group::_zoom_in_callback() { zoom_in_callback(true); }
void Group::_zoom_out_callback() { zoom_out_callback(true); }
void Group::_zoom_reset_callback() { zoom_reset_callback(true); }

void Group::_reset_callback() { reset_sig.emit(true); }

void Group::fit_callback(bool) { fit_sig.emit(true); }
void Group::_fit_callback() { fit_callback(true); }

void Group::grid(GRID in)
{
  if (in == _grid) return;
  
  _grid = in;
  
  overlay_sig.emit(true);
  update_sig.emit(true);
}

GRID Group::grid() const { return _grid; }

void Group::grid_callback(GRID in) { grid(in); }

void Group::_grid_callback()
{
  grid(GRID(list::find(_p->menu->item(), _p->menu_grid)));
}

void Group::hud(bool in)
{
  if (in == _hud) return;
  
  _hud = in;
  
  overlay_sig.emit(true);
  update_sig.emit(true);
}

bool Group::hud() const { return _hud; }

void Group::hud_callback(bool in) { hud(in); }
void Group::_hud_callback() { hud(_p->menu->value()); }

}
}

