/*------------------------------------------------------------------------------
 include/djv_base/mtx_etc_.h

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include <djv_base/math.h>

namespace djv_base {
namespace mtx {

template<typename T, int D>
inline Mtx<T, D> transpose(const Mtx<T, D> & in)
{
  Mtx<T, D> out;
  for (int i = 0; i < D; ++i)
    for (int j = 0; j < D; ++j)
      out.e[j * D + i] = in.e[i * D + j];
  return out;
}

template<typename T, int D>
inline Mtx<T, D> operator * (const Mtx<T, D> & a, const Mtx<T, D> & b)
{
  Mtx<T, D> out;
  out.zero();
  for (int i = 0; i < D; ++i)
    for (int j = 0; j < D; ++j)
      for (int k = 0; k < D; ++k)
        out.e[i * D + j] += a.e[i * D + k] * b.e[k * D + j];
  return out;
}

template<typename T, int D, int D2>
inline Mtx<T, D> translate(const Vec<T, D2> & in)
{
  //DJV_ASSERT((D - 1) == D2);
  
  Mtx<T, D> out;
  for (int i = 0; i < D2; ++i)
    out.e[(D - 1) * D + i] = in.e[i];
  return out;
}

inline Mtx3f translate3f(const Vec2f & in)
{
  return translate<double, 3, 2>(in);
}

template<typename T, int D, int D2>
inline Mtx<T, D> scale(const Vec<T, D2> & in)
{
  //DJV_ASSERT((D - 1) == D2);
  
  Mtx<T, D> out;
  for (int i = 0; i < D2; ++i)
    out.e[i * D + i] = in.e[i];
  return out;
}

inline Mtx3f scale3f(const Vec2f & in)
{
  return scale<double, 3, 2>(in);
}

inline Mtx3f rotate3f(double in)
{
  const double cos = math::cos(math::degrees_to_radians(in));
  const double sin = math::sin(math::degrees_to_radians(in));

  return Mtx3f(
    cos,  sin, 0,
    -sin, cos, 0,
    0,    0,   1);
}

template<typename T, int D, int D2>
inline Vec<T, D2> operator * (const Mtx<T, D> & a, const Vec<T, D2> & b)
{
  //DJV_ASSERT((D - 1) == D2);

  Vec<T, D2> out;
  for (int i = 0; i < D2; ++i)
  {
    int j = 0;
    for (; j < D2; ++j)
      out.e[i] += a.e[j * D + i] * b.e[j];
    out.e[i] += a.e[j * D + i];
  }
  return out;
}

template<typename T>
inline Box<T, 2> operator * (const Mtx<T, 3> & a, const Box<T, 2> & b)
{
  const Vec<T, 2> position1 = b.position1();
  Vec<T, 2> pt [] = { b.position, b.position, position1, b.position };
  pt[1].y = position1.y;
  pt[3].x = position1.x;
  
  Box<T, 2> out;
  for (int i = 0; i < 4; ++i)
  {
    pt[i] = a * pt[i];
    if (0 == i) out.position = pt[i];
    else out = box::expand(out, pt[i]);
  }
  return out;
}

template<typename T, int D>
inline bool operator == (const Mtx<T, D> & a, const Mtx<T, D> & b)
{
  for (int i = 0; i < a.dimension_x2; ++i)
    if (a.e[i] != b.e[i]) return false;
  return true;
}

template<typename T, int D>
inline bool operator != (const Mtx<T, D> & a, const Mtx<T, D> & b)
{
  return ! (a == b);
}

template <typename T, typename T2, int D>
inline Mtx<T2, D> convert(const Mtx<T, D> & in)
{
  Mtx<T2, D> out;
  for (int i = 0; i < out.dimension_x2; ++i)
    out.e[i] = T2(in.e[i]);
  return out;
}

// XXX Probably wrong.

template <typename T>
Mtx<T, 4> mtx4(const Mtx<T, 3> & in)
{
  return Mtx<T, 4>(
    in.e[0], in.e[1], in.e[2], 0.0,
    in.e[3], in.e[4], in.e[5], 0.0,
    0.0,     0.0,     in.e[8], 0.0,
    in.e[6], in.e[7], 0.0,     1.0);
}

template<typename T, int D>
inline Str & operator << (Str & out, const Mtx<T, D> & in)
{
  for (int i = 0; i < in.dimension_x2; ++i)
    out << in.e[i];
  return out;
}

template<typename T, int D>
inline Str & operator >> (Str & in, Mtx<T, D> & out) throw (Str)
{
  for (int i = 0; i < out.dimension_x2; ++i)
    in >> out.e[i];
  return in;
}

template<typename T, int D>
inline Dbg & operator << (Dbg & dbg, const Mtx<T, D> & in)
{
  dbg << "\n";
  for (int i = 0; i < D; ++i)
  {
    for (int j = 0; j < D; ++j)
    {
      dbg << in.e[i * D + j];
      if (j < D - 1) dbg << " ";
    }
    if (i < D - 1) dbg << "\n";
  }
  return dbg;
}

}
}

