/*------------------------------------------------------------------------------
 include/djv_image/gl.h
 
 OpenGL
 ------
 
 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_GL_H
#define DJV_IMAGE_GL_H

#include <djv_image/pixel.h>
#include <djv_base/box.h>
#include <djv_base/vec.h>

namespace djv_image {
namespace gl {

//------------------------------------------------------------------------------
// Options

enum PIPELINE
{
  FIXED,
  PROGRAM
};

//------------------------------------------------------------------------------
// Functions

void ortho(const Vec2i & size, const Vec2f & min_max = Vec2f(-1.0, 1.0));

GLenum format(PIXEL, bool bgr = false);
GLenum type(PIXEL);
GLenum texture();
  
inline void box(const Box2i &);
inline void box(const Box2f &);
inline void box(const Box2i &, const Vec2f uv [4]);
inline void box(const Box2f &, const Vec2f uv [4]);

//------------------------------------------------------------------------------
// Conversion

Str & operator >> (Str &, PIPELINE &) throw (Str);
Str & operator << (Str &, PIPELINE);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_pipeline();

//------------------------------------------------------------------------------
// Debugging

Dbg & operator << (Dbg &, PIPELINE);

// References:
//
// * Dave Shreiner, "Performance OpenGL: Platform Independent Techniques",
//   SIGGRAPH 2001.

#if defined(DJV_DBG)
#define DJV_DBG_GL(cmd) \
  \
  cmd; \
  { \
    ::GLenum error = GL_NO_ERROR; \
    if ((error = ::glGetError()) != GL_NO_ERROR) \
    { \
      DJV_DBG_PRN(str::Format("%% = %% (%%, line %%)"). \
        arg(#cmd).arg((char *)::gluErrorString(error)). \
        arg(__FILE__).arg(__LINE__)); \
    } \
  }
#else
#define DJV_DBG_GL(cmd) cmd;
#endif

//------------------------------------------------------------------------------
// Global

class Device
{
public:

  virtual ~Device() = 0;

  virtual void init() throw (Error) = 0;

  const Str & vendor_str() const;
  const Str & render_str() const;
  const Str & version_str() const;

protected:

  Str _vendor_str, _render_str, _version_str;
};

class Global
{
public:
  
  Global() throw (Error);
  ~Global();
  
  Device * device() const;

  void pipeline(PIPELINE);
  PIPELINE pipeline() const;
  
private:

  std::auto_ptr<Device> _device;
  PIPELINE _pipeline;
};

Global * global();

}
}

#include <djv_image/gl_.h>

#endif

