/*------------------------------------------------------------------------------
 include/djv_image/gl_data.h
 
 OpenGL Image Data
 -----------------

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_GL_DATA_H
#define DJV_IMAGE_GL_DATA_H

#include <djv_image/color.h>
#include <djv_image/data.h>
#include <djv_image/gl.h>
#include <djv_image/gl_offscreen.h>
#include <djv_image/profile.h>
#include <djv_base/box.h>
#include <djv_base/mtx.h>

namespace djv_image {
namespace gl_data {

//------------------------------------------------------------------------------
// Options

struct Xform
{
  Xform() :
    scale(1.0),
    rotate(0.0)
  {}

  Vec2b mirror;
  Vec2f position;
  Vec2f scale;
  double rotate;
};

enum CHANNEL
{
  CHANNEL_DEFAULT,
  CHANNEL_0,
  CHANNEL_1,
  CHANNEL_2,
  CHANNEL_3
};
  
struct Color
{
  Color() :
    brightness(1),
    contrast(1),
    saturation(1)
  {}

  double brightness, contrast, saturation;
};

struct Levels
{
  Levels() :
    in_low(0.0),
    in_high(1.0),
    gamma(1.0),
    out_low(0.0),
    out_high(1.0)
  {}

  double in_low, in_high;
  double gamma;
  double out_low, out_high;
};

struct Display
{
  Display() :
    soft_clip(0.0)
  {}

  djv_image::Data lut;
  Color color;
  Levels levels;
  double soft_clip;
};

struct Filter
{
  enum FILTER
  {
    FILTER_DEFAULT,
    FILTER_CUSTOM
  };

  enum DEFAULT
  {
    LINEAR,
    NEAREST
  };

  enum CUSTOM
  {
    BOX,
    TRIANGLE,
    BELL,
    BSPLINE,
    LANCZOS3,
    CUBIC,
    MITCHELL
  };
  
  typedef std::pair<DEFAULT, DEFAULT> Default;
  typedef std::pair<CUSTOM, CUSTOM> Custom;

  Filter();
  
  FILTER value;
  Default _default;
  Custom custom;
};

struct Option
{
  Option() :
    channel(CHANNEL(0)),
    clamp(false)
  {}

  Xform xform;
  Profile profile;
  CHANNEL channel;
  Display display;
  Filter filter;
  djv_image::Color background;
  
  // XXX NV_float_buffer doesn't clamp glReadPixels, so we have to do it
  // manually for integer formats.
  //
  // References:
  //
  // * http://oss.sgi.com/projects/ogl-sample/registry/NV/float_buffer.txt
  
  bool clamp;
};

//------------------------------------------------------------------------------
// Data

class Data
{
public:

  Data();
  ~Data();

  // Draw pixels.
  
  void draw(
    const djv_image::Data &,
    const Option & = Option()
  ) const throw (Error);
  
  // Read pixels.
  
  static void read(djv_image::Data *);
  static void read(djv_image::Data *, const Box2i &);
  
private:

  Option _option;

  struct _Data;
  std::auto_ptr<_Data> _p;
};

//------------------------------------------------------------------------------
// Functions

// OpenGL state.

void state_unpack(const djv_image::Info &, const Vec2i & offset = Vec2i());
void state_pack(const djv_image::Info &, const Vec2i & offset = Vec2i());
void state_reset();

// Copy.

void copy(
  const djv_image::Data &,
  djv_image::Data *,
  const Option & = Option(),
  Data * = 0,
  gl_offscreen::Buffer * = 0
) throw (Error);

// Transform.

Mtx3f xform_mtx(const Xform &);

// Color.

Mtx4f brightness_mtx(double r, double g, double b);
Mtx4f contrast_mtx(double r, double g, double b);
Mtx4f saturation_mtx(double r, double g, double b);

Mtx4f color_mtx(const Color &);
djv_image::Data color_lut(const Levels &, double soft_clip);

// Average.

void average(
  const djv_image::Data &,
  djv_image::Color *
) throw (Error);

// Histogram.

enum HISTOGRAM
{
  HISTOGRAM_256,
  HISTOGRAM_1024
};

void histogram(
  const djv_image::Data &,
  djv_image::Data *,
  HISTOGRAM,
  djv_image::Color * min,
  djv_image::Color * max
) throw (Error);

int histogram_size(HISTOGRAM);

//------------------------------------------------------------------------------
// Comparison

bool operator == (const Xform &, const Xform &);
bool operator == (const Color &, const Color &);
bool operator == (const Levels &, const Levels &);
bool operator == (const Filter &, const Filter &);
bool operator != (const Xform &, const Xform &);
bool operator != (const Color &, const Color &);
bool operator != (const Levels &, const Levels &);
bool operator != (const Filter &, const Filter &);

//------------------------------------------------------------------------------
// Conversion

GLuint filter_to_gl(Filter::DEFAULT);

Str & operator >> (Str &, Xform &) throw (Str);
Str & operator >> (Str &, Color &) throw (Str);
Str & operator >> (Str &, Levels &) throw (Str);
Str & operator >> (Str &, Filter &) throw (Str);
Str & operator >> (Str &, Filter::Default &) throw (Str);
Str & operator >> (Str &, Filter::Custom &) throw (Str);
Str & operator >> (Str &, Filter::FILTER &) throw (Str);
Str & operator >> (Str &, Filter::DEFAULT &) throw (Str);
Str & operator >> (Str &, Filter::CUSTOM &) throw (Str);
Str & operator >> (Str &, CHANNEL &) throw (Str);
Str & operator >> (Str &, HISTOGRAM &) throw (Str);
Str & operator << (Str &, const Xform &);
Str & operator << (Str &, const Color &);
Str & operator << (Str &, const Levels &);
Str & operator << (Str &, const Filter &);
Str & operator << (Str &, const Filter::Default &);
Str & operator << (Str &, const Filter::Custom &);
Str & operator << (Str &, Filter::FILTER);
Str & operator << (Str &, Filter::DEFAULT);
Str & operator << (Str &, Filter::CUSTOM);
Str & operator << (Str &, CHANNEL);
Str & operator << (Str &, HISTOGRAM);

//------------------------------------------------------------------------------
// Labels

const List<Str> & label_channel();
const List<Str> & label_filter();
const List<Str> & label_filter_default();
const List<Str> & label_filter_custom();
const List<Str> & label_histogram();

//------------------------------------------------------------------------------
// Global

class Global
{
public:

  Global();
  ~Global();

  void filter(Filter::FILTER);
  void filter_default(const Filter::Default &);
  void filter_custom(const Filter::Custom &);
  Filter::FILTER filter() const;
  const Filter::Default & filter_default() const;
  const Filter::Custom & filter_custom() const;

private:

  Filter::FILTER _filter;
  Filter::Default _filter_default;
  Filter::Custom _filter_custom;
};

Global * global();

//------------------------------------------------------------------------------
// Debugging

Dbg & operator << (Dbg &, const Xform &);
Dbg & operator << (Dbg &, CHANNEL);
Dbg & operator << (Dbg &, const Color &);
Dbg & operator << (Dbg &, const Levels &);
Dbg & operator << (Dbg &, const Filter &);
Dbg & operator << (Dbg &, const Filter::Default &);
Dbg & operator << (Dbg &, const Filter::Custom &);
Dbg & operator << (Dbg &, HISTOGRAM);

}
}

#endif

