/*------------------------------------------------------------------------------
 include/djv_image/io.h
 
 Image I/O
 ---------
 
 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#ifndef DJV_IMAGE_IO_H
#define DJV_IMAGE_IO_H

#include <djv_image/image.h>
#include <djv_image/time.h>
#include <djv_base/file.h>
#include <djv_base/io.h>
#include <djv_base/plugin.h>

namespace djv_image {
namespace io {

//------------------------------------------------------------------------------
// Plugin Base

class Plugin : public plugin::Plugin
{
public:
  
  // List of supported file extensions.
  
  virtual const List<Str> & extension_list() const { return _null_list; }
  
  // Plugin options.
  
  virtual bool option(const Str &, Str *) { return false; }
  virtual Str option(const Str &) const { return Str(); }
  virtual const List<Str> & option_list() const { return _null_list; }

private:

  //static const List<Str> _null_list; // XXX MinGW/gcc 3.4.2 bug?
  List<Str> _null_list;
};

//------------------------------------------------------------------------------
// Load Plugin Base

class Load : public Plugin
{
public:
  
  // Open.
  
  virtual void open(
    const File &,
    Info *,
    Tag *,
    Time *
  ) throw (Error) = 0;
  
  virtual void open(
    const File &,
    List<Info> *,
    Tag *,
    Time *
  ) throw (Error) = 0;

  // Load.
  
  virtual const Image * get(
    int frame = -1,
    int layer = 0,
    data::PROXY = data::PROXY_NONE
  ) throw (Error) = 0;

protected:

  // Utility class for frame-based file formats.

  struct Cache
  {
    Cache();
    Cache(const File & file, int frame, int layer, data::PROXY);

    File file;
    int frame;
    int layer;
    data::PROXY proxy;
    
    bool operator == (const Cache &) const;
  };
};

//------------------------------------------------------------------------------
// Save Plugin Base

class Save : public Plugin
{
public:

  virtual ~Save() { close(); }

  // Open.
  
  virtual void open(
    const File &,
    const Info &,
    const Tag &,
    const Time &
  ) throw (Error) = 0;
  
  // Close.
  
  virtual void close() throw (Error) {}
  
  // Save.
  
  virtual void set(
    const Image &,
    int frame = -1
  ) throw (Error) = 0;
};

//------------------------------------------------------------------------------
// Load Plugin Factory

class Load_Factory : public plugin::Factory
{
public:

  Load_Factory();
  ~Load_Factory();
  
  // Test if a file is an image.
  
  bool valid(const File &, const Str & plugin = Str()) const;

  // Find a plugin for the given file.
  
  Load * get(
    const File &,
    Info * = 0,
    Tag * = 0,
    Time * = 0,
    const Str & plugin = Str()
  ) const throw (Error);

  Load * get(
    const File &,
    List<Info> *,
    Tag * = 0,
    Time * = 0,
    const Str & plugin = Str()
  ) const throw (Error);
  
  // Plugin options.

  bool option(const Str & plugin, const Str &, Str *);
  
  // Plugin list.
  
  List<Load *> list() const;

private:

  Str plugin_entry() const { return "djv_image_load"; }
};

//------------------------------------------------------------------------------
// Save Plugin Factory

class Save_Factory : public plugin::Factory
{
public:

  Save_Factory();
  ~Save_Factory();
  
  // Test if a file is an image.
  
  bool valid(const File &, const Str & plugin = Str());

  // Find a plugin for the given file.
  
  Save * get(
    const File &,
    const Info &,
    const Tag & tag = Tag(),
    const Time & = Time(),
    const Str & plugin = Str()
  ) const throw (Error);

  // Plugin options.

  bool option(const Str & plugin, const Str &, Str *);
  
  // Plugin list.
  
  List<Save *> list() const;

private:

  Str plugin_entry() const { return "djv_image_save"; }
};

//------------------------------------------------------------------------------
// Labels

extern const Str
  error,
  error_unrecognized,
  error_unsupported,
  error_open,
  error_read,
  error_write;

//------------------------------------------------------------------------------
// Global

class Global
{
public:

  Global();
  ~Global();

  bool valid(const File &, const Str & plugin = Str()) const;

  Load * load(
    const File &,
    Info * = 0,
    Tag * = 0,
    Time * = 0,
    const Str & plugin = Str()
  ) const throw (Error);

  Load * load(
    const File &,
    List<Info> *,
    Tag * = 0,
    Time * = 0,
    const Str & plugin = Str()
  ) const throw (Error);

  Save * save(
    const File &,
    const Info &,
    const Tag & = Tag(),
    const Time & = Time(),
    const Str & plugin = Str()
  ) const throw (Error);

  Load_Factory * load() const;
  Save_Factory * save() const;

private:

  std::auto_ptr<Load_Factory> _load;
  std::auto_ptr<Save_Factory> _save;
};

Global * global();

}
}

#endif

