/*------------------------------------------------------------------------------
 lib/djv_base/io.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG
//#undef DJV_MMAP

#include <djv_base/file.h>
#include <djv_base/io.h>
#if defined(DJV_MMAP)
#if ! defined(WIN32)
#include <sys/mman.h>
#endif // WIN32
#endif // DJV_MMAP
#if defined(WIN32)
#include <io.h>
#else // WIN32
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#endif // WIN32
#include <errno.h>

namespace djv_base {
namespace io {

Io::Io() :
#if defined(WIN32)
  _f(INVALID_HANDLE_VALUE),
  _mode(MODE(0)),
  _size(0),
  _endian(false),
  _mmap(0), _mmap_start(0), _mmap_end(0), _mmap_p(0)
#else
  _f(-1),
  _mode(MODE(0)),
  _size(0),
  _endian(false),
  _mmap((void *)-1), _mmap_start(0), _mmap_end(0), _mmap_p(0)
#endif
{}

Io::~Io()
{
  close();
}

void Io::open(const Str & in, MODE mode) throw (Error)
{
  DJV_DBG_OBJ("Io::open");
  DJV_DBG_PRN("in = " << in);
  DJV_DBG_PRN("mode = " << mode);

  close();

  // Open file.
  
#if defined(WIN32)

  _f = ::CreateFile(
    in.c_str(),
    (WRITE == mode) ? GENERIC_WRITE : GENERIC_READ,
    (WRITE == mode) ? 0 : FILE_SHARE_READ,
    0,
    (WRITE == mode) ? CREATE_ALWAYS : OPEN_EXISTING,
    //FILE_ATTRIBUTE_NORMAL,
    //FILE_FLAG_NO_BUFFERING |
    FILE_FLAG_SEQUENTIAL_SCAN,
    0);

  if (INVALID_HANDLE_VALUE == _f)
    throw Error(error, error_open);

#else // WIN32

  int read_flag = 0;
#if defined(DJV_LINUX)
  //read_flag = O_DIRECT;
#endif
  
  _f = ::open(
    in.c_str(),
    (WRITE == mode) ? (O_WRONLY | O_CREAT | O_TRUNC) : (O_RDONLY | read_flag),
    (WRITE == mode) ? (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH) : (0));

  if (-1 == _f)
    throw Error(error, error_open);

#endif // WIN32

  // File size.
  
#if defined(WIN32)
  _size = ::GetFileSize(_f, 0);
#else
  _size = File(in).size;
#endif
  DJV_DBG_PRN("size = " << _size);

  // I/O mode.
  
  _mode = mode;
  
  // Memory maping.
  
#if defined(DJV_MMAP)

  if (READ == _mode)
  {
    DJV_DBG_PRN("mmap");
  
#if defined(WIN32)

    if (! (_mmap = ::CreateFileMapping(_f, 0, PAGE_READONLY, 0, 0, 0)))
      throw Error(error, error_mmap);
    
    _mmap_start = reinterpret_cast<const uint8_t *>(
      ::MapViewOfFile(_mmap, FILE_MAP_READ, 0, 0, 0));
    
    if (! _mmap_start)
      throw Error(error, error_mmap);

    _mmap_end = _mmap_start + _size;
    _mmap_p = _mmap_start;

#else // WIN32

    if (
      ((_mmap = ::mmap(0, _size, PROT_READ, MAP_SHARED, _f, 0)) == (void *)-1)
    )
      throw Error(error, error_mmap);

    _mmap_start = reinterpret_cast<const uint8_t *>(_mmap);
    _mmap_end = _mmap_start + _size;
    _mmap_p = _mmap_start;

#endif // WIN32
  }
#endif // DJV_MMAP
}

void Io::close()
{
  DJV_DBG_OBJ("Io::close");

#if defined(DJV_MMAP)
#if defined(WIN32)

  if (_mmap_start != 0)
  {
    ::UnmapViewOfFile((void *)_mmap_start);
    _mmap_start = 0;
  }
  if (_mmap != 0) { ::CloseHandle(_mmap); _mmap = 0; }

#else // WIN32

  _mmap_start = 0;
  if (_mmap != (void *)-1)
  {
    DJV_DBG_PRN("munmap");
    
    // XXX Solaris wants (char *)?
    
    int r = ::munmap((char *)_mmap, _size);
    if (-1 == r)
    {
      const Str err(strerror(errno));
      DJV_DBG_PRN("errno = " << err);
    }
    _mmap = (void *)-1;
  }

#endif // WIN32

  _mmap_end = 0;
  _mmap_p = 0;

#endif // DJV_MMAP

#if defined(WIN32)
  if (_f != INVALID_HANDLE_VALUE)
  {
    ::CloseHandle(_f);
    _f = INVALID_HANDLE_VALUE;
  }
#else // WIN32
  if (_f != -1)
  {
    DJV_DBG_PRN("close");
    
    int r = ::close(_f);
    if (-1 == r)
    {
      const Str err(strerror(errno));
      DJV_DBG_PRN("errno = " << err);
    }
    _f = -1;
  }
#endif // WIN32

  _size = 0;
  _mode = MODE(0);
}
  
bool Io::valid() const
{
  return
#if defined(WIN32)
    _f != INVALID_HANDLE_VALUE &&
#else
    _f != -1 &&
#endif
    position() < off_t(_size);
}

void Io::set(const void * in, size_t size, size_t word_size) throw (Error)
{
  DJV_DBG_OBJ("Io::set");
  DJV_DBG_PRN("size = " << size);
  DJV_DBG_PRN("word size = " << word_size);
  DJV_DBG_PRN("endian = " << _endian);
  
  uint8_t * p = (uint8_t *)in;
  
  mem::Buffer<uint8_t> tmp;
  
  if (_endian && word_size > 1)
  {
    tmp.size(size * word_size);
    p = tmp();
    mem::endian(in, p, size, word_size);
  }

#if defined(WIN32)
  DWORD n;
  if (! ::WriteFile(_f, p, size * word_size, &n, 0))
    throw Error(error, error_write);
#else
  if (::write(_f, p, size * word_size) == -1)
    throw Error(error, error_write);
#endif
}

void Io::set_8(const int8_t * in, size_t size) throw (Error)
  { return set(in, size, 1); }
void Io::set_u8(const uint8_t * in, size_t size) throw (Error)
  { return set(in, size, 1); }
void Io::set_16(const int16_t * in, size_t size) throw (Error)
  { return set(in, size, 2); }
void Io::set_u16(const uint16_t * in, size_t size) throw (Error)
  { return set(in, size, 2); }
void Io::set_32(const int32_t * in, size_t size) throw (Error)
  { return set(in, size, 4); }
void Io::set_u32(const uint32_t * in, size_t size) throw (Error)
  { return set(in, size, 4); }
void Io::set_f32(const float * in, size_t size) throw (Error)
  { return set(in, size, 4); }

void Io::get(void * in, size_t size, size_t word_size) throw (Error)
{
  DJV_DBG_OBJ("Io::get");
  DJV_DBG_PRN("size = " << size);
  DJV_DBG_PRN("word size = " << word_size);

#if defined(DJV_MMAP)

  const uint8_t * p = _mmap_p + size * word_size;
  if (p > _mmap_end)
    throw Error(error, error_read);
  
  if (_endian && word_size > 1)
    mem::endian(_mmap_p, in, size, word_size);
  else
    mem::copy(_mmap_p, in, size * word_size);

  _mmap_p = p;

#else // DJV_MMAP

#if defined(WIN32)
  DWORD n;
  if (! ::ReadFile(_f, in, size * word_size, &n, 0))
    throw Error(error, error_read);
#else
  //if (-1 == ::read(_f, in, size * word_size) == (size * word_size))
  //  throw Error(error, error_read);
  ::read(_f, in, size * word_size);
  
  if (_endian && word_size > 1)
    mem::endian(in, size, word_size);
#endif

#endif // DJV_MMAP
}

void Io::get_8(int8_t * in, size_t size) throw (Error)
  { return get(in, size, 1); }
void Io::get_u8(uint8_t * in, size_t size) throw (Error)
  { return get(in, size, 1); }
void Io::get_16(int16_t * in, size_t size) throw (Error)
  { return get(in, size, 2); }
void Io::get_u16(uint16_t * in, size_t size) throw (Error)
  { return get(in, size, 2); }
void Io::get_32(int32_t * in, size_t size) throw (Error)
  { return get(in, size, 4); }
void Io::get_u32(uint32_t * in, size_t size) throw (Error)
  { return get(in, size, 4); }
void Io::get_f32(float * in, size_t size) throw (Error)
  { return get(in, size, 4); }

void Io::readahead()
{
#if defined(DJV_MMAP)

#if defined(DJV_LINUX)

  ::madvise((void *)_mmap_start, _size, MADV_WILLNEED);

#elif defined(DJV_SOLARIS)

  ::madvise((caddr_t)_mmap_start, _size, MADV_WILLNEED);

#else

  // XXX Force readahead?

  for (const uint8_t * p = _mmap_start; p < _mmap_end; p += mem::align)
    _mmap_readahead = *p;

#endif

#else // DJV_MMAP

#if defined(DJV_LINUX)

  ::posix_fadvise(_f, 0, _size, POSIX_FADV_NOREUSE);
  ::posix_fadvise(_f, 0, _size, POSIX_FADV_WILLNEED);

#endif

#endif // DJV_MMAP
}

void Io::position(off_t in) throw (Error)
{
  position(in, false);
}

void Io::seek(off_t in) throw (Error)
{
  position(in, true);
}

void Io::position(off_t in, bool seek) throw (Error)
{
  DJV_DBG_OBJ("Io::position");
  DJV_DBG_PRN("in = " << int(in));
  DJV_DBG_PRN("seek = " << seek);
  
  switch (_mode)
  {
    case READ:
#if defined(DJV_MMAP)
      if (! seek)
        _mmap_p = reinterpret_cast<const uint8_t *>(_mmap_start) + in;
      else
        _mmap_p += in;
      
      if (_mmap_p > _mmap_end)
        throw Error(error, error_position);
      break;
#else
      // Fall through...
#endif

    case WRITE:
    {
#if defined(WIN32)
      if (! ::SetFilePointer(_f, in, 0, ! seek ? FILE_BEGIN : FILE_CURRENT))
        throw Error(error, error_position);
#else
      const off_t p = ::lseek(_f, in, ! seek ? SEEK_SET : SEEK_CUR);

      if (p == (off_t)-1)
        throw Error(error, error_position);
#endif
    } break;
  }
}

off_t Io::position() const
{
  off_t out = 0;
  
  switch (_mode)
  {
    case READ:
#if defined(DJV_MMAP)
      out = off_t(_mmap_p - reinterpret_cast<const uint8_t *>(_mmap_start));
      break;
#else
    // Fall through...
#endif

    case WRITE:
#if defined(WIN32)
      out = ::SetFilePointer(_f, 0, 0, FILE_CURRENT);
#else
      out = ::lseek(_f, 0, SEEK_CUR);
#endif
      break;
  }

  return out;
}

void Io::endian(bool in)
{
  _endian = in;
}

}
}

