/*------------------------------------------------------------------------------
 lib/djv_image/app.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include <djv_image/app.h>
#include <FL/Fl.H>

namespace djv_image {
namespace app {

App::App(const Str & name, int & argc, char ** argv) throw (Error) :
  djv_base::app::App(name, argc, argv)
{
  DJV_DBG_OBJ("App::App");
  
  Fl::visual(FL_DOUBLE | FL_RGB);
  
  // Global.
  
  _gl = std::auto_ptr<gl::Global>(new gl::Global);
  _gl_data = std::auto_ptr<gl_data::Global>(new gl_data::Global);
  _gl_offscreen = std::auto_ptr<gl_offscreen::Global>(new gl_offscreen::Global);
  _time = std::auto_ptr<time::Global>(new time::Global);
  _io = std::auto_ptr<io::Global>(new io::Global);
}

App::~App()
{
  DJV_DBG_OBJ("App::~App");
}
  
gl::Global * App::gl() const { return _gl.get(); }
gl_data::Global * App::gl_data() const { return _gl_data.get(); }
gl_offscreen::Global * App::gl_offscreen() const { return _gl_offscreen.get(); }
time::Global * App::time() const { return _time.get(); }
io::Global * App::io() const { return _io.get(); }

Str App::info() const
{
  return str::Format(label_info).
    arg(djv_base::app::App::info()).
    arg(str::label(gl::global()->pipeline())).
    arg(str::label(gl_offscreen::global()->buffer())).
    arg(gl::global()->device()->vendor_str()).
    arg(gl::global()->device()->render_str()).
    arg(gl::global()->device()->version_str()).
    arg(io::global()->load()->name_list(), ", ").
    arg(io::global()->save()->name_list(), ", ");
}

void App::cmdln(Str & in) throw (Error)
{
  DJV_DBG_OBJ("App::cmdln");
  DJV_DBG_PRN("in = " << in);
  
  djv_base::app::App::cmdln(in);
  
  const List<io::Load *> & load = io::global()->load()->list();
  const List<io::Save *> & save = io::global()->save()->list();

  Str tmp, arg;
try
{
  while (! in.empty())
  {
    in >> arg;
    
    // Time.
    
    if ("-time_units" == arg)
    {
      time::UNITS value = time::UNITS(0);
      in >> value;
      time::global()->units(value);
    }
    else if ("-speed_default" == arg)
    {
      time::FPS value = time::FPS(0);
      in >> value;
      time::global()->speed(value);
    }
    
    // OpenGL.
    
    else if ("-gl_pipeline" == arg)
    {
      gl::PIPELINE value = gl::PIPELINE(0);
      in >> value;
      gl::global()->pipeline(value);
    }
    else if ("-gl_filter" == arg)
    {
      gl_data::Filter::FILTER value = gl_data::Filter::FILTER(0);
      in >> value;
      gl_data::global()->filter(value);
    }
    else if ("-gl_filter_default" == arg)
    {
      gl_data::Filter::Default value;
      in >> value;
      gl_data::global()->filter_default(value);
    }
    else if ("-gl_filter_custom" == arg)
    {
      gl_data::Filter::Custom value;
      in >> value;
      gl_data::global()->filter_custom(value);
    }
    else if ("-gl_offscreen" == arg)
    {
      gl_offscreen::BUFFER value = gl_offscreen::BUFFER(0);
      in >> value;
      gl_offscreen::global()->buffer(value);
    }
    
    // Image I/O.
    
    else if ("-io_load" == arg)
    {
      Str plugin;
      in >> plugin;
      for (uint i = 0; i < load.size(); ++i)
        if (str::compare(plugin, load[i]->name()))
        {
          Str option;
          if (! str::serialize(&in, &option, load[i]->option_list())) throw in;
          if (! load[i]->option(option, &in)) throw in;
          break;
        }
    }
    else if ("-io_save" == arg)
    {
      Str plugin;
      in >> plugin;
      for (uint i = 0; i < save.size(); ++i)
        if (str::compare(plugin, save[i]->name()))
        {
          Str option;
          if (! str::serialize(&in, &option, save[i]->option_list())) throw in;
          if (! save[i]->option(option, &in)) throw in;
          break;
        }
    }
    
    else tmp << arg;
  }
}
catch (Str) { throw Error(arg); }
  
  in = tmp;
}

Str App::cmdln_help() const
{
  return str::Format(label_cmdln_help).
    arg(time::label_units(), ", ").
    arg(str::label(time::global()->units())).
    arg(time::label_fps(), ", ").
    arg(str::label(time::global()->speed())).
    arg(gl::label_pipeline(), ", ").
    arg(str::label(gl::global()->pipeline())).
    arg(gl_data::label_filter(), ", ").
    arg(str::label(gl_data::global()->filter())).
    arg(gl_data::label_filter_default(), ", ").
    arg(str::label(gl_data::global()->filter_default())).
    arg(gl_data::label_filter_custom(), ", ").
    arg(str::label(gl_data::global()->filter_custom())).
    arg(gl_offscreen::label_buffer(), ", ").
    arg(str::label(gl_offscreen::global()->buffer())).
    arg(djv_base::app::App::cmdln_help());
}

}
}
