/*------------------------------------------------------------------------------
 lib/djv_image/gl_data_etc.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG0
//#define DJV_DBG

#include <djv_image/gl_data.h>

namespace djv_image {
namespace gl_data {

void state_unpack(const Info & in, const Vec2i & offset)
{
  glPixelStorei(GL_UNPACK_ALIGNMENT, in.align);
  glPixelStorei(GL_UNPACK_SWAP_BYTES, in.endian != mem::endian());
  glPixelStorei(GL_UNPACK_ROW_LENGTH, in.size.x);
  glPixelStorei(GL_UNPACK_SKIP_ROWS, offset.y);
  glPixelStorei(GL_UNPACK_SKIP_PIXELS, offset.x);
}

void state_pack(const Info & in, const Vec2i & offset)
{
  glPixelStorei(GL_PACK_ALIGNMENT, in.align);
  glPixelStorei(GL_PACK_SWAP_BYTES, in.endian != mem::endian());
  glPixelStorei(GL_PACK_ROW_LENGTH, in.size.x);
  glPixelStorei(GL_PACK_SKIP_ROWS, offset.y);
  glPixelStorei(GL_PACK_SKIP_PIXELS, offset.x);
}

void state_reset()
{
  state_pack(Info());
  state_unpack(Info());
  if (GLEW_ARB_shader_objects) glUseProgramObjectARB(0);
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
}

Mtx3f xform_mtx(const Xform & in)
{
  return
    mtx::scale3f(in.scale) *
    mtx::rotate3f(in.rotate) *
    mtx::translate3f(in.position);
}

Mtx4f brightness_mtx(double r, double g, double b)
{
  return Mtx4f(
    r, 0, 0, 0,
    0, g, 0, 0,
    0, 0, b, 0,
    0, 0, 0, 1);
}

Mtx4f contrast_mtx(double r, double g, double b)
{
  return Mtx4f(
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    -0.5, -0.5, -0.5, 1) *
    Mtx4f(
    r, 0, 0, 0,
    0, g, 0, 0,
    0, 0, b, 0,
    0, 0, 0, 1) *
    Mtx4f(
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    0.5, 0.5, 0.5, 1);
}

Mtx4f saturation_mtx(double r, double g, double b)
{
  const float s [] =
  {
    (1.0 - r) * 0.3086,
    (1.0 - g) * 0.6094,
    (1.0 - b) * 0.0820
  };
  return Mtx4f(
    s[0] + r, s[0],     s[0],     0,
    s[1],     s[1] + g, s[1],     0,
    s[2],     s[2],     s[2] + b, 0,
    0,        0,        0,        1);
}

Mtx4f color_mtx(const Color & in)
{
  return
    brightness_mtx(in.brightness, in.brightness, in.brightness) *
    contrast_mtx(in.contrast, in.contrast, in.contrast) *
    saturation_mtx(in.saturation, in.saturation, in.saturation);
}

djv_image::Data color_lut(const Levels & in, double soft_clip)
{
  djv_image::Data out(Info(Vec2i(1024, 1), pixel::L_F32));
  
  const double in_tmp = in.in_high - in.in_low;
  const double gamma = 1.0 / in.gamma;
  const double out_tmp = in.out_high - in.out_low;

  pixel::F32_T * p = reinterpret_cast<pixel::F32_T *>(out.data());
  const int size = out.size().x;
  for (int i = 0; i < size; ++i, ++p)
    *p = math::soft_clip(
      math::pow(
        math::max(((i / double(size - 1) - in.in_low) / in_tmp), 0.000001),
        gamma
      ) * out_tmp + in.out_low,
      soft_clip);
  
  return out;
}

bool operator == (const Xform & a, const Xform & b)
{
  return
    a.mirror == b.mirror &&
    a.position == b.position &&
    a.scale == b.scale &&
    a.rotate == b.rotate;
}

bool operator != (const Xform & a, const Xform & b)
{
  return ! (a == b);
}

bool operator == (const Color & a, const Color & b)
{
  return
    a.brightness == b.brightness &&
    a.contrast == b.contrast &&
    a.saturation == b.saturation;
}

bool operator != (const Color & a, const Color & b)
{
  return ! (a == b);
}

bool operator == (const Levels & a, const Levels & b)
{
  return
    a.in_low == b.in_low &&
    a.in_high == b.in_high &&
    a.gamma == b.gamma &&
    a.out_low == b.out_low &&
    a.out_high == b.out_high;
}

bool operator != (const Levels & a, const Levels & b)
{
  return ! (a == b);
}

bool operator == (const Filter & a, const Filter & b)
{
  return
    a.value == b.value &&
    a._default == b._default &&
    a.custom == b.custom;
}

bool operator != (const Filter & a, const Filter & b)
{
  return ! (a == b);
}

GLuint filter_to_gl(Filter::DEFAULT in)
{
  switch (in)
  {
    case Filter::LINEAR: return GL_LINEAR;
    case Filter::NEAREST: return GL_NEAREST;
  }
  return 0;
}

Str & operator >> (Str & in, Xform & out) throw (Str)
{
  return in >>
    out.mirror >>
    out.position >>
    out.scale >>
    out.rotate;
}

Str & operator << (Str & out, const Xform & in)
{
  return out <<
    in.mirror <<
    in.position <<
    in.scale <<
    in.rotate;
}

Str & operator >> (Str & in, Color & out) throw (Str)
{
  return in >>
    out.brightness >>
    out.contrast >>
    out.saturation;
}

Str & operator << (Str & out, const Color & in)
{
  return out <<
    in.brightness <<
    in.contrast <<
    in.saturation;
}

Str & operator >> (Str & in, Levels & out) throw (Str)
{
  return in >>
    out.in_low >>
    out.in_high >>
    out.gamma >>
    out.out_low >>
    out.out_high;
}

Str & operator << (Str & out, const Levels & in)
{
  return out <<
    in.in_low <<
    in.in_high <<
    in.gamma <<
    in.out_low <<
    in.out_high;
}

Str & operator >> (Str & in, Filter & out) throw (Str)
{
  return in >>
    out.value >>
    out._default >>
    out.custom;
}

Str & operator << (Str & out, const Filter & in)
{
  return out <<
    in.value <<
    in._default <<
    in.custom;
}

Str & operator >> (Str & in, Filter::Default & out) throw (Str)
{
  return in >>
    out.first >>
    out.second;
}

Str & operator << (Str & out, const Filter::Custom & in)
{
  return out <<
    in.first <<
    in.second;
}

Str & operator >> (Str & in, Filter::Custom & out) throw (Str)
{
  return in >>
    out.first >>
    out.second;
}

Str & operator << (Str & out, const Filter::Default & in)
{
  return out <<
    in.first <<
    in.second;
}

DJV_STR_OPERATOR_SET(Filter::FILTER, label_filter())
DJV_STR_OPERATOR_GET(Filter::FILTER, label_filter())
DJV_STR_OPERATOR_SET(Filter::DEFAULT, label_filter_default())
DJV_STR_OPERATOR_GET(Filter::DEFAULT, label_filter_default())
DJV_STR_OPERATOR_SET(Filter::CUSTOM, label_filter_custom())
DJV_STR_OPERATOR_GET(Filter::CUSTOM, label_filter_custom())
DJV_STR_OPERATOR_SET(CHANNEL, label_channel())
DJV_STR_OPERATOR_GET(CHANNEL, label_channel())
DJV_STR_OPERATOR_SET(HISTOGRAM, label_histogram())
DJV_STR_OPERATOR_GET(HISTOGRAM, label_histogram())

const List<Str> & label_channel()
{
  static const List<Str> data = List<Str>() <<
    "Default" <<
    "0" <<
    "1" <<
    "2" <<
    "3";
  return data;
}

const List<Str> & label_filter()
{
  static const List<Str> data = List<Str>() <<
    "Default" <<
    "Custom";
  return data;
}

const List<Str> & label_filter_default()
{
  static const List<Str> data = List<Str>() <<
    "Linear" <<
    "Nearest";
  return data;
}

const List<Str> & label_filter_custom()
{
  static const List<Str> data = List<Str>() <<
    "Box" <<
    "Triangle" <<
    "Bell" <<
    "B-Spline" <<
    "Lanczos3" <<
    "Cubic" <<
    "Mitchell";
  return data;
}

Global::Global() :
  _filter(Filter::FILTER_DEFAULT),
  _filter_default(Filter::LINEAR, Filter::LINEAR),
  _filter_custom(Filter::LANCZOS3, Filter::MITCHELL)
{}

Global::~Global()
{}

void Global::filter(Filter::FILTER in)
{
  DJV_DBG_OBJ("Global::filter");
  DJV_DBG_PRN("in = " << in);
  
  _filter = in;
}

Filter::FILTER Global::filter() const { return _filter; }

void Global::filter_default(const Filter::Default & in)
{
  _filter_default = in;
}

const Filter::Default & Global::filter_default() const
  { return _filter_default; }

void Global::filter_custom(const Filter::Custom & in)
{
  _filter_custom = in;
}

const Filter::Custom & Global::filter_custom() const
  { return _filter_custom; }

Dbg & operator << (Dbg & dbg, const Xform & in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, CHANNEL in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, const Color & in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, const Levels & in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, const Filter & in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, const Filter::Default & in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, const Filter::Custom & in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, Filter::FILTER in)
  { return dbg << str::label(in); }
Dbg & operator << (Dbg & dbg, HISTOGRAM in)
  { return dbg << str::label(in); }

}
}
