/*------------------------------------------------------------------------------
 lib/djv_image/glx.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#if ! defined(WIN32) && ! defined(DJV_MACOSX)

//#define DJV_DBG

#include "glx.h"
#include <djv_base/math.h>
#include <FL/x.H>

namespace djv_image {
namespace gl {

_Glx_Device::_Glx_Device() :
  _window(0),
  _context(0)
{
  DJV_DBG_OBJ("_Glx_Device::_Glx_Device");
}

_Glx_Device::~_Glx_Device()
{
  DJV_DBG_OBJ("_Glx_Device::~_Glx_Device");

  if (_context)
  {
    DJV_DBG_PRN("context");
    glXDestroyContext(fl_display, _context);
  }
  
  if (_window)
  {
    DJV_DBG_PRN("window");
    XDestroyWindow(fl_display, _window);
  }
}

void _Glx_Device::init() throw (Error)
{
  DJV_DBG_OBJ("_Glx_Device::init");
  
  // Open X display.
  
  fl_open_display();
  
  // Check for GLX support.
  
  if (! glXQueryExtension(fl_display, 0, 0))
    throw Error(error_glx, error_glx_init);

  // XXX Create a dummy window and rendering context for glewInit.
  // According to the docs, glewInit can be called just once per-process?
  
  XSetWindowAttributes win_attrib;
  win_attrib.colormap = fl_colormap;
  win_attrib.border_pixel = 0;
  _window = XCreateWindow(
    fl_display,
    RootWindow(fl_display, fl_screen),
    0, 0, 1, 1, 0,
    fl_visual->depth,
    InputOutput,
    fl_visual->visual,
    CWColormap | CWBorderPixel,
    &win_attrib);
  if (! _window)
    throw Error(error_x11, error_x11_window);

  // Create context.
  
  _context = glXCreateContext(fl_display, fl_visual, 0, True);
  if (! _context)
    throw Error(error_glx, error_glx_context_create);
  
  if (! glXMakeCurrent(fl_display, _window, _context))
    throw Error(error_glx, error_glx_context_bind);
  
  // Initialize GLEW.
  
  GLint error = glewInit();
  if (error != GLEW_OK)
    throw Error(error_glew, str::Format(error_glew_init).
      arg((char *)glewGetErrorString(error)));

  _vendor_str = Str((const char *)glGetString(GL_VENDOR));
  _render_str = Str((const char *)glGetString(GL_RENDERER));
  _version_str = Str((const char *)glGetString(GL_VERSION));
  DJV_DBG_PRN(
    "direct rendering = " << glXIsDirect(fl_display, _context));
  DJV_DBG_PRN(
    "server glx vendor string = " <<
    glXQueryServerString(fl_display, fl_screen, GLX_VENDOR));
  DJV_DBG_PRN(
    "server glx version string = " <<
    glXQueryServerString(fl_display, fl_screen, GLX_VERSION));
  DJV_DBG_PRN(
    "server glx extensions = " <<
    glXQueryServerString(fl_display, fl_screen, GLX_EXTENSIONS));
  DJV_DBG_PRN(
   "client glx vendor string = " <<
    glXGetClientString(fl_display, GLX_VENDOR));
  DJV_DBG_PRN(
    "client glx version string = " <<
    glXGetClientString(fl_display, GLX_VERSION));
  DJV_DBG_PRN(
    "client glx extensions = " <<
    glXGetClientString(fl_display, GLX_EXTENSIONS));
  DJV_DBG_PRN(
    "GLX extensions = " <<
    glXQueryExtensionsString(fl_display, fl_screen));
  DJV_DBG_PRN("OpenGL vendor string = " << _vendor_str);
  DJV_DBG_PRN("OpenGL renderer string = " << _render_str);
  DJV_DBG_PRN("OpenGL version string = " << _version_str);
  DJV_DBG_PRN(
    "OpenGL extensions = " << (const char *)glGetString(GL_EXTENSIONS));
  DJV_DBG_PRN("glu version = " << (const char *)gluGetString(GLU_VERSION));
  DJV_DBG_PRN(
    "glu extensions = " << (const char *)gluGetString(GLU_EXTENSIONS));
}

Global::Global() throw (Error) :
  _device(new _Glx_Device),
  _pipeline(FIXED)
{
  DJV_DBG_OBJ("Global::Global");
  
  _device->init();
  
  const Str & version = _device->version_str();
  if (version.size())
    if ('2' == version[0])
      _pipeline = PROGRAM;
}

Global::~Global()
{}

}
}

#endif
