/*------------------------------------------------------------------------------
 lib/djv_image/io_cineon_info.cxx
 
 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io_cineon_info.h"

namespace djv_image {
namespace io_cineon {
namespace info {

Header::Header()
{
  mem::set(0xff, &_file, sizeof(File));
  zero(_file.version, 8);
  zero(_file.name, 100);
  zero(_file.time, 24);

  mem::set(0xff, &_image, sizeof(Image));
  for (uint i = 0; i < 8; ++i)
  {
    zero(&_image.channel[i].low_data);
    zero(&_image.channel[i].low_quantity);
    zero(&_image.channel[i].high_data);
    zero(&_image.channel[i].high_quantity);
  }

  mem::set(0xff, &_source, sizeof(Source));
  zero(&_source.offset[0]);
  zero(&_source.offset[1]);
  zero(_source.file, 100);
  zero(_source.time, 24);
  zero(_source.input_device, 64);
  zero(_source.input_model, 32);
  zero(_source.input_serial, 32);
  zero(&_source.input_pitch[0]);
  zero(&_source.input_pitch[1]);
  zero(&_source.gamma);

  mem::set(0xff, &_film, sizeof(Film));
  zero(_film.format, 32);
  zero(&_film.frame_rate);
  zero(_film.frame_id, 32);
  zero(_film.slate, 200);
}

void Header::load(djv_base::io::Io * io, Info * info, Tag * tag,
  time::Speed * speed, bool * print) throw (Error)
{
  DJV_DBG_OBJ("Header::load");
  
  // Read.
  
  io->get(&_file, sizeof(File));
  
  bool endian = false;
  if (magic[0] == _file.magic)
    ;
  else if (magic[1] == _file.magic)
    endian = true;
  else
    throw Error(name, io::error_unrecognized);

  io->get(&_image, sizeof(Image));
  io->get(&_source, sizeof(Source));
  io->get(&_film, sizeof(Film));

  if (endian)
  {
    DJV_DBG_PRN("endian");
    io->endian(true);
    this->endian();
    info->endian = mem::endian_opposite(mem::endian());
  }
  
  // Information.
  
  if (_file.image_offset)
    io->position(_file.image_offset);

  switch (_image.orient)
  {
    case ORIENT_LEFT_RIGHT_TOP_BOTTOM: info->mirror.y = true; break;
    case ORIENT_LEFT_RIGHT_BOTTOM_TOP: break;
    case ORIENT_RIGHT_LEFT_TOP_BOTTOM: info->mirror.x = true; break;
    case ORIENT_RIGHT_LEFT_BOTTOM_TOP:
      info->mirror.x = info->mirror.y = true; break;
    case ORIENT_TOP_BOTTOM_LEFT_RIGHT:
    case ORIENT_TOP_BOTTOM_RIGHT_LEFT:
    case ORIENT_BOTTOM_TOP_LEFT_RIGHT:
    case ORIENT_BOTTOM_TOP_RIGHT_LEFT: break;
  }
  
  if (! _image.channels)
    throw Error(name, io::error_unsupported);
  int i = 1;
  for (; i < _image.channels; ++i)
  {
    if (
      (_image.channel[i].size[0] != _image.channel[0].size[0]) ||
      (_image.channel[i].size[1] != _image.channel[0].size[1])
    )
      break;
    if (_image.channel[i].bit_depth != _image.channel[0].bit_depth)
      break;
  }
  if (i < _image.channels)
    throw Error(name, io::error_unsupported);
  
  int pixel = -1;
  switch (_image.channels)
  {
    case 3:
      switch (_image.channel[0].bit_depth)
      {
        case 10: pixel = pixel::RGB_U10; break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(name, io::error_unsupported);
  info->pixel = PIXEL(pixel);
  
  info->size = Vec2i(_image.channel[0].size[0], _image.channel[0].size[1]);
      
  if (valid(&_image.line_padding) && _image.line_padding)
    throw Error(name, io::error_unsupported);
  if (valid(&_image.channel_padding) && _image.channel_padding)
    throw Error(name, io::error_unsupported);
  
  if (print)
    *print = DESCRIPTOR_R_PRINT == _image.channel[0].descriptor[1];
  
  // File image tags.
  
  const List<Str> & label = tag::label_tag();
  const List<Str> & label_cineon = io_cineon::label_tag();

  if (valid(_file.time, 24))
    (*tag)[label[tag::TIME]] = to_str(_file.time, 24);
    
  // Source image tags.
  
  if (valid(&_source.offset[0]) && valid(&_source.offset[1]))
    (*tag)[label_cineon[TAG_SOURCE_OFFSET]] = str::join(List<Str>() <<
      str::int_to_str(_source.offset[0]) <<
      str::int_to_str(_source.offset[1]),
      " ");

  if (valid(_source.file, 100))
    (*tag)[label_cineon[TAG_SOURCE_FILE]] = to_str(_source.file, 100);
  
  if (valid(_source.time, 24))
    (*tag)[label_cineon[TAG_SOURCE_TIME]] = to_str(_source.time, 24);
  
  if (valid(_source.input_device, 64))
    (*tag)[label_cineon[TAG_SOURCE_INPUT_DEVICE]] =
      to_str(_source.input_device, 64);
  
  if (valid(_source.input_model, 32))
    (*tag)[label_cineon[TAG_SOURCE_INPUT_MODEL]] =
      to_str(_source.input_model, 32);
  
  if (valid(_source.input_serial, 32))
    (*tag)[label_cineon[TAG_SOURCE_INPUT_SERIAL]] =
      to_str(_source.input_serial, 32);
  
  if (valid(&_source.input_pitch[0]) && valid(&_source.input_pitch[1]))
    (*tag)[label_cineon[TAG_SOURCE_INPUT_PITCH]] = str::join(List<Str>() <<
      str::float_to_str(_source.input_pitch[0]) <<
      str::float_to_str(_source.input_pitch[1]),
      " ");
   
  if (valid(&_source.gamma))
    (*tag)[label_cineon[TAG_SOURCE_GAMMA]] = str::float_to_str(_source.gamma);
      
  // Film image tags.
  
  if (
    valid(&_film.id) &&
    valid(&_film.type) &&
    valid(&_film.offset) &&
    valid(&_film.prefix) &&
    valid(&_film.count)
  )
    (*tag)[label[tag::KEYCODE]] = time::keycode_to_str(
      _film.id, _film.type, _film.prefix, _film.count, _film.offset);

  if (valid(_film.format, 32))
    (*tag)[label_cineon[TAG_FILM_FORMAT]] = to_str(_film.format, 32);
  
  if (valid(&_film.frame))
    (*tag)[label_cineon[TAG_FILM_FRAME]] = str::int_to_str(_film.frame);
  
  if (valid(&_film.frame_rate) && _film.frame_rate >= speed_min)
  {
    if (speed) *speed = time::float_to_speed(_film.frame_rate);
    (*tag)[label_cineon[TAG_FILM_FRAME_RATE]] =
      str::float_to_str(_film.frame_rate);
  }
  
  if (valid(_film.frame_id, 32))
    (*tag)[label_cineon[TAG_FILM_FRAME_ID]] = to_str(_film.frame_id, 32);
  
  if (valid(_film.slate, 200))
    (*tag)[label_cineon[TAG_FILM_SLATE]] = to_str(_film.slate, 200);
  
  // End.
  
  dbg();
}

void Header::save(djv_base::io::Io * io, const Str & file, const Info & info,
  const Tag & tag, COLOR color) throw (Error)
{
  DJV_DBG_OBJ("Header::save");

  // Information.
  
  _file.image_offset = 2048;
  _file.header_size = 1024;
  _file.industry_header_size = 1024;
  _file.user_header_size = 0;
  
  _image.orient = ORIENT_LEFT_RIGHT_TOP_BOTTOM;

  _image.channels = 3;  
  if (COLOR_PRINT == color)
  {
    _image.channel[0].descriptor[1] = DESCRIPTOR_R_PRINT;
    _image.channel[1].descriptor[1] = DESCRIPTOR_G_PRINT;
    _image.channel[2].descriptor[1] = DESCRIPTOR_B_PRINT;
  }
  else
  {
    _image.channel[0].descriptor[1] = DESCRIPTOR_L;
    _image.channel[1].descriptor[1] = DESCRIPTOR_L;
    _image.channel[2].descriptor[1] = DESCRIPTOR_L;
  }
  
  const int bit_depth = 10;
  for (int i = 0; i < _image.channels; ++i)
  {
    _image.channel[i].descriptor[0] = 0;
    _image.channel[i].bit_depth = bit_depth;
    _image.channel[i].size[0] = info.size.x;
    _image.channel[i].size[1] = info.size.y;

    _image.channel[i].low_data = 0;
    switch (bit_depth)
    {
      case 8: _image.channel[i].high_data = 255; break;
      case 10: _image.channel[i].high_data = 1023; break;
      case 12: _image.channel[i].high_data = 4095; break;
      case 16: _image.channel[i].high_data = 65535; break;
    }
  }
  
  _image.interleave = 0;
  _image.packing = 5;
  _image.data_sign = 0;
  _image.data_sense = 0;
  _image.line_padding = 0;
  _image.channel_padding = 0;
  
  // File image tags.

  const List<Str> & label = tag::label_tag();
  const List<Str> & label_cineon = io_cineon::label_tag();
  Str tmp;
  
  str::cstr(file, _file.name, 100, false);
  str::cstr(tag[label[tag::TIME]], _file.time, 24, false);
  
  // Source image tags.
  
  tmp = tag[label_cineon[TAG_SOURCE_OFFSET]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, ' ');
    if (2 == list.size())
    {
      _source.offset[0] = str::str_to_int(list[0]);
      _source.offset[1] = str::str_to_int(list[1]);
    }
  }
  
  tmp = tag[label_cineon[TAG_SOURCE_FILE]];
  if (tmp.size()) str::cstr(tmp, _source.file, 100, false);

  tmp = tag[label_cineon[TAG_SOURCE_TIME]];
  if (tmp.size()) str::cstr(tmp, _source.time, 24, false);

  tmp = tag[label_cineon[TAG_SOURCE_INPUT_DEVICE]];
  if (tmp.size()) str::cstr(tmp, _source.input_device, 64, false);

  tmp = tag[label_cineon[TAG_SOURCE_INPUT_MODEL]];
  if (tmp.size()) str::cstr(tmp, _source.input_model, 32, false);

  tmp = tag[label_cineon[TAG_SOURCE_INPUT_SERIAL]];
  if (tmp.size()) str::cstr(tmp, _source.input_serial, 32, false);
  
  tmp = tag[label_cineon[TAG_SOURCE_INPUT_PITCH]];
  if (tmp.size())
  {
    const List<Str> list = str::split(tmp, ' ');
    if (2 == list.size())
    {
      _source.offset[0] = str::str_to_int(list[0]);
      _source.offset[1] = str::str_to_int(list[1]);
    }
  }
  
  tmp = tag[label_cineon[TAG_SOURCE_GAMMA]];
  if (tmp.size()) _source.gamma = str::str_to_float(tmp);
  
  // Film image tags.

  tmp = tag[label[tag::KEYCODE]];
  if (tmp.size())
  {
    int id = 0, type = 0, prefix = 0, count = 0, offset = 0;
    time::str_to_keycode(tmp, &id, &type, &prefix, &count, &offset);
    _film.id = id;
    _film.type = type;
    _film.offset = offset;
    _film.prefix = prefix;
    _film.count = count;
  }
  
  tmp = tag[label_cineon[TAG_FILM_FORMAT]];
  if (tmp.size()) str::cstr(tmp, _film.format, 32, false);

  tmp = tag[label_cineon[TAG_FILM_FRAME]];
  if (tmp.size()) _film.frame = str::str_to_int(tmp);

  tmp = tag[label_cineon[TAG_FILM_FRAME_RATE]];
  if (tmp.size()) _film.frame_rate = str::str_to_float(tmp);

  tmp = tag[label_cineon[TAG_FILM_FRAME_ID]];
  if (tmp.size()) str::cstr(tmp, _film.frame_id, 32, false);

  tmp = tag[label_cineon[TAG_FILM_SLATE]];
  if (tmp.size()) str::cstr(tmp, _film.slate, 200, false);
  
  // Write.
  
  dbg();
  
  const bool endian = mem::endian() != mem::MSB;
  io->endian(endian);
  if (endian)
  {
    DJV_DBG_PRN("endian");
    this->endian();
    _file.magic = magic[1];
  }
  else
  {
    _file.magic = magic[0];
  }
  
  io->set(&_file, sizeof(File));
  io->set(&_image, sizeof(Image));
  io->set(&_source, sizeof(Source));
  io->set(&_film, sizeof(Film));
}

void Header::endian()
{
  mem::endian(&_file.image_offset, 1, 4);
  mem::endian(&_file.header_size, 1, 4);
  mem::endian(&_file.industry_header_size, 1, 4);
  mem::endian(&_file.user_header_size, 1, 4);
  mem::endian(&_file.size, 1, 4);

  for (uint i = 0; i < 8; ++i)
  {
    mem::endian(_image.channel[i].size, 2, 4);
    mem::endian(&_image.channel[i].low_data, 1, 4);
    mem::endian(&_image.channel[i].low_quantity, 1, 4);
    mem::endian(&_image.channel[i].high_data, 1, 4);
    mem::endian(&_image.channel[i].high_quantity, 1, 4);
  }
  mem::endian(_image.white, 2, 4);
  mem::endian(_image.red, 2, 4);
  mem::endian(_image.green, 2, 4);
  mem::endian(_image.blue, 2, 4);
  mem::endian(&_image.line_padding, 1, 4);
  mem::endian(&_image.channel_padding, 1, 4);
  
  mem::endian(_source.offset, 2, 4);
  mem::endian(_source.input_pitch, 2, 4);
  mem::endian(&_source.gamma, 1, 4);
  
  mem::endian(&_film.prefix, 1, 4);
  mem::endian(&_film.count, 1, 4);
  mem::endian(&_film.frame, 1, 4);
  mem::endian(&_film.frame_rate, 1, 4);
}

void Header::dbg() const
{
  using info::dbg;
  
  DJV_DBG_OBJ("dbg");
  
  DJV_DBG_PRN("file");
  DJV_DBG_PRN("  image offset = " << dbg(_file.image_offset));
  DJV_DBG_PRN("  header size = " << dbg(_file.header_size));
  DJV_DBG_PRN("  industry header size = " << dbg(_file.industry_header_size));
  DJV_DBG_PRN("  user header size = " << dbg(_file.user_header_size));
  DJV_DBG_PRN("  size = " << dbg(_file.size));
  DJV_DBG_PRN("  version = " << dbg(_file.version, 8));
  DJV_DBG_PRN("  name = " << dbg(_file.name, 100));
  DJV_DBG_PRN("  time = " << dbg(_file.time, 24));

  DJV_DBG_PRN("image");
  DJV_DBG_PRN("  orient = " << dbg(_image.orient));
  DJV_DBG_PRN("  channels = " << dbg(_image.channels));
  for (int i = 0; i < _image.channels; ++i)
  {
    DJV_DBG_PRN("  channel " << i);
    DJV_DBG_PRN("    descriptor = " <<
      dbg(_image.channel[i].descriptor[0]) << " " <<
      dbg(_image.channel[i].descriptor[1]));
    DJV_DBG_PRN("    bit depth = " << dbg(_image.channel[i].bit_depth));
    DJV_DBG_PRN("    size = " << dbg(_image.channel[i].size[0]) << " " <<
      dbg(_image.channel[i].size[1]));
    DJV_DBG_PRN("    low data = " << dbg(_image.channel[i].low_data));
    DJV_DBG_PRN("    low quantity = " << dbg(_image.channel[i].low_quantity));
    DJV_DBG_PRN("    high data = " << dbg(_image.channel[i].high_data));
    DJV_DBG_PRN("    high quantity = " <<
      dbg(_image.channel[i].high_quantity));
  }
  DJV_DBG_PRN("  white = " << dbg(_image.white[0]) << " " <<
    dbg(_image.white[1]));
  DJV_DBG_PRN("  red = " << dbg(_image.red[0]) << " " <<
    dbg(_image.red[1]));
  DJV_DBG_PRN("  green = " << dbg(_image.green[0]) << " " <<
    dbg(_image.green[1]));
  DJV_DBG_PRN("  blue = " << dbg(_image.blue[0]) << " " <<
    dbg(_image.blue[1]));
  DJV_DBG_PRN("  label = " << dbg(_image.label, 200));
  DJV_DBG_PRN("  interleave = " << dbg(_image.interleave));
  DJV_DBG_PRN("  packing = " << dbg(_image.packing));
  DJV_DBG_PRN("  data sign = " << dbg(_image.data_sign));
  DJV_DBG_PRN("  data sense = " << dbg(_image.data_sense));
  DJV_DBG_PRN("  line padding = " << dbg(_image.line_padding));
  DJV_DBG_PRN("  channel padding = " << dbg(_image.channel_padding));
  
  DJV_DBG_PRN("source");
  DJV_DBG_PRN("  offset = " << dbg(_source.offset[0]) << " " <<
    dbg(_source.offset[1]));
  DJV_DBG_PRN("  file = " << dbg(_source.file, 100));
  DJV_DBG_PRN("  time = " << dbg(_source.time, 24));
  DJV_DBG_PRN("  input device = " << dbg(_source.input_device, 64));
  DJV_DBG_PRN("  input model = " << dbg(_source.input_model, 32));
  DJV_DBG_PRN("  input serial = " << dbg(_source.input_serial, 32));
  DJV_DBG_PRN("  input pitch = " << dbg(_source.input_pitch[0]) << " " <<
    dbg(_source.input_pitch[1]));
  DJV_DBG_PRN("  gamma = " << dbg(_source.gamma));

  DJV_DBG_PRN("film");
  DJV_DBG_PRN("  id = " << dbg(_film.id));
  DJV_DBG_PRN("  type = " << dbg(_film.type));
  DJV_DBG_PRN("  offset = " << dbg(_film.offset));
  DJV_DBG_PRN("  prefix = " << dbg(_film.prefix));
  DJV_DBG_PRN("  count = " << dbg(_film.count));
  DJV_DBG_PRN("  format = " << dbg(_film.format, 32));
  DJV_DBG_PRN("  frame = " << dbg(_film.frame));
  DJV_DBG_PRN("  frame rate = " << dbg(_film.frame_rate));
  DJV_DBG_PRN("  frame id = " << dbg(_film.frame_id, 32));
  DJV_DBG_PRN("  slate = " << dbg(_film.slate, 200));
}

}

void info_load(djv_base::Io * io, Info * info, Tag * tag, time::Speed * speed,
  bool * print) throw (Error)
{
  info::Header header;
  header.load(io, info, tag, speed, print);
}

void info_save(djv_base::Io * io, const Str & file, const Info & info,
  const Tag & tag, COLOR color) throw (Error)
{
  info::Header header;
  header.save(io, file, info, tag, color);
}

void info_update(djv_base::Io * io) throw (Error)
{
  const size_t size = io->position();
  
  io->position(20);
  io->set_u32(size);
}

}
}
