/*------------------------------------------------------------------------------
 lib/djv_image/io_tga_info.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#include <djv_image/io_tga.h>

namespace djv_image {
namespace io_tga {

namespace {

struct Header
{
  Header();

  void load(djv_base::io::Io *, Info *, bool * compress) throw (Error);
  void save(djv_base::io::Io *, const Info &, bool compress)
    throw (Error);
  
  void dbg() const;

private:

  struct Data
  {
    uint8_t  id_size;
    uint8_t  cmap_type;
    uint8_t  image_type;
    uint16_t cmap_start;
    uint16_t cmap_size;
    uint8_t  cmap_bits;
    uint16_t x;
    uint16_t y;
    uint16_t width;
    uint16_t height;
    uint8_t  pixel_bits;
    uint8_t  descriptor;
  } _data;
};

Header::Header()
{
  _data.id_size = 0;
  _data.cmap_type = 0;
  _data.image_type = 0;
  _data.cmap_start = 0;
  _data.cmap_size = 0;
  _data.cmap_bits = 0;
  _data.x = 0;
  _data.y = 0;
  _data.width = 0;
  _data.height = 0;
  _data.pixel_bits = 0;
  _data.descriptor = 0;
}

void Header::load(djv_base::io::Io * in, Info * info, bool * compress)
  throw (Error)
{
  DJV_DBG_OBJ("Header::load");
  
  // Read.
  
  in->get_u8(&_data.id_size);
  in->get_u8(&_data.cmap_type);
  in->get_u8(&_data.image_type);
  in->get_u16(&_data.cmap_start);
  in->get_u16(&_data.cmap_size);
  in->get_u8(&_data.cmap_bits);
  in->get_u16(&_data.x);
  in->get_u16(&_data.y);
  in->get_u16(&_data.width);
  in->get_u16(&_data.height);
  in->get_u8(&_data.pixel_bits);
  in->get_u8(&_data.descriptor);

  in->seek(_data.id_size);
  
  // Information.

  info->size = Vec2i(_data.width, _data.height);
  
  info->mirror = Vec2b(
    (_data.descriptor >> 4) & 1,
    (_data.descriptor >> 5) & 1);

  const int alpha_bits = _data.descriptor & 15;
  DJV_DBG_PRN("alpha bits = " << alpha_bits);
  switch (alpha_bits)
  {
    case 0:
    case 8: break;
    default: throw Error(name, io::error_unsupported);
  }

  int pixel = -1;
  switch (_data.image_type)
  {
    case 3:
    case 11:
      switch (_data.pixel_bits)
      {
        case 8: pixel = pixel::L_U8; break;
        case 16: pixel = pixel::LA_U8; break;
      }
      break;

    case 2:
    case 10:
      switch (_data.pixel_bits)
      {
        case 24: pixel = pixel::RGB_U8; info->bgr = true; break;
        case 32: pixel = pixel::RGBA_U8; info->bgr = true; break;
      }
      break;
  }
  if (-1 == pixel)
    throw Error(name, io::error_unsupported);
  info->pixel = PIXEL(pixel);
  DJV_DBG_PRN("pixel = " << info->pixel);
  
  const int bits = _data.pixel_bits + alpha_bits;
  if (bits < (pixel::channels(info->pixel) * 8) || (bits % 8) != 0)
    throw Error(name, io::error_unsupported);
  
  *compress =
    10 == _data.image_type ||
    11 == _data.image_type;
  DJV_DBG_PRN("compress = " << compress);
  
  info->endian = mem::LSB;
}

void Header::save(djv_base::io::Io * in, const Info & info, bool compress)
  throw (Error)
{
  DJV_DBG_OBJ("Header::save");
  
  // Information.
  
  _data.width = info.size.x;
  _data.height = info.size.y;

  const int channels = pixel::channels(info.pixel);
  _data.pixel_bits = channels * 8;
  switch (channels)
  {
    case 1:
    case 2: _data.image_type = 3; break;
    case 3:
    case 4: _data.image_type = 2; break;
  }
  
  if (compress)
    _data.image_type += 8;
  
  int alpha_bits = 0;
  switch (pixel::format(info.pixel))
  {
    case pixel::LA:
    case pixel::RGBA: alpha_bits = 8; break;
    default: break;
  }
  
  const int orient = 0;
  
  _data.descriptor = (alpha_bits & 15) | ((orient & 3) << 4);

  // Write.
  
  in->set_u8(_data.id_size);
  in->set_u8(_data.cmap_type);
  in->set_u8(_data.image_type);
  in->set_u16(_data.cmap_start);
  in->set_u16(_data.cmap_size);
  in->set_u8(_data.cmap_bits);
  in->set_u16(_data.x);
  in->set_u16(_data.y);
  in->set_u16(_data.width);
  in->set_u16(_data.height);
  in->set_u8(_data.pixel_bits);
  in->set_u8(_data.descriptor);
}

void Header::dbg() const
{
  DJV_DBG_OBJ("Header::dbg");
  DJV_DBG_PRN("id size = " << _data.id_size);
  DJV_DBG_PRN("cmap type = " << _data.cmap_type);
  DJV_DBG_PRN("image type = " << _data.image_type);
  DJV_DBG_PRN("cmap start = " << _data.cmap_start);
  DJV_DBG_PRN("cmap size = " << _data.cmap_size);
  DJV_DBG_PRN("cmap bits = " << _data.cmap_bits);
  DJV_DBG_PRN("image position = " <<_data. x << " " << _data.y);
  DJV_DBG_PRN("image size = " << _data.width << " " << _data.height);
  DJV_DBG_PRN("pixel bits = " << _data.pixel_bits);
}

}

void info_load(djv_base::io::Io * io, Info * info, bool * compress) throw (Error)
{
  Header header;
  header.load(io, info, compress);
}

void info_save(djv_base::io::Io * io, const Info & info, bool compress)
  throw (Error)
{
  Header header;
  header.save(io, info, compress);
}

}
}
