/*------------------------------------------------------------------------------
 lib/djv_image/wgl.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#if defined(WIN32)

//#define DJV_DBG

#include "wgl.h"

namespace djv_image {
namespace gl {

_Wgl_Device::_Wgl_Device() :
  _id(0),
  _device(0),
  _context(0)
{
  DJV_DBG_OBJ("_Wgl_Device::_Wgl_Device");
}

_Wgl_Device::~_Wgl_Device()
{
  DJV_DBG_OBJ("_Wgl_Device::~_Wgl_Device");
  
  if (_context)
  {
    DJV_DBG_PRN("context");
    wglDeleteContext(_context);
  }
  
  if (_device)
  {
    DJV_DBG_PRN("device");
    ReleaseDC(_id, _device);
  }
}

void _Wgl_Device::init() throw (Error)
{
  DJV_DBG_OBJ("_Wgl_Device::init");

  // XXX Create a dummy window and rendering context for glewInit.
  // According to the docs, glewInit can be called just once per-process?
  
  HINSTANCE hinstance = GetModuleHandle(0);
  if (! hinstance)
    throw Error(error_win32, str::Format(error_win32_module).
      arg(int(GetLastError())));
  
  static const char name [] = "djv";
  WNDCLASS wc;
  if (! GetClassInfo(hinstance, name, &wc))
  {
    wc.style = CS_OWNDC;
    //wc.lpfnWndProc = (WNDPROC)MainWndProc;
    wc.lpfnWndProc = DefWindowProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = hinstance;
    wc.hIcon = LoadIcon(0, IDI_APPLICATION);
    wc.hCursor = LoadCursor(0, IDC_ARROW);
    wc.hbrBackground = 0;
    wc.lpszMenuName = 0;
    wc.lpszClassName = name;
    
    if (! RegisterClass(&wc))
      throw Error(error_win32, str::Format(error_win32_register).
        arg(int(GetLastError())));
  }
  
  _id = CreateWindow(name, 0, 0, 0, 0, 0, 0, 0, 0, hinstance, 0);
  if (! _id)
    throw Error(error_win32, str::Format(error_win32_window).
      arg(int(GetLastError())));
  
  _device = GetDC(_id);
  if (! _device)
    throw Error(error_win32, str::Format(error_win32_device).
      arg(int(GetLastError())));
  
  PIXELFORMATDESCRIPTOR pixel_format_info;
  const int pixel_format_size = DescribePixelFormat(_device, 0, 0, 0);
  DJV_DBG_PRN("pixel format size = " << pixel_format_size);
  for (int i = 1; i < pixel_format_size; ++i)
  {
    DescribePixelFormat(_device, i, sizeof(PIXELFORMATDESCRIPTOR),
      &pixel_format_info);
    DJV_DBG_PRN("  id " << i << ": " <<
      ((PFD_SUPPORT_OPENGL & pixel_format_info.dwFlags) ? "ogl " : "") <<
      ((PFD_GENERIC_FORMAT & pixel_format_info.dwFlags) ? "" : "accel ") <<
      ((PFD_TYPE_RGBA == pixel_format_info.iPixelType) ? "rgba " : "") <<
      "depth = " << pixel_format_info.cColorBits << "/" <<
        pixel_format_info.cRedBits << "/" <<
        pixel_format_info.cGreenBits << "/" <<
        pixel_format_info.cBlueBits << "/" <<
        pixel_format_info.cAlphaBits << " ");
  }
  
  PIXELFORMATDESCRIPTOR pixel_format =
  {
    sizeof(PIXELFORMATDESCRIPTOR),
    1,
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL,
    PFD_TYPE_RGBA,
    32,
    0, 0, 0, 0, 0, 0,
    0,
    0,
    0,
    0, 0, 0, 0,
    0,
    0,
    0,
    PFD_MAIN_PLANE,
    0,
    0, 0, 0
  };
  int pixel_format_id = ChoosePixelFormat(_device, &pixel_format);
  DJV_DBG_PRN("pixel format = " << pixel_format_id);
  if (! pixel_format_id)
    throw Error(error_win32, str::Format(error_win32_pixel_choose).
      arg(int(GetLastError())));
  
  if (! SetPixelFormat(_device, pixel_format_id, &pixel_format))
    throw Error(error_win32, str::Format(error_win32_pixel_set).
      arg(int(GetLastError())));
  
  // Create context.
  
  _context = wglCreateContext(_device);
  if (! _context)
    throw Error(error_wgl, str::Format(error_wgl_context_create).
      arg(int(GetLastError())));
  
  if (! wglMakeCurrent(_device, _context))
    throw Error(error_wgl, str::Format(error_wgl_context_bind).
      arg(int(GetLastError())));
  
  // Initialize GLEW.
  
  GLenum err = glewInit();
  if (err != GLEW_OK)
    throw Error(error_glew, str::Format(error_glew_init).
      arg((char *)glewGetErrorString(err)));

  _vendor_str = Str((const char *)glGetString(GL_VENDOR));
  _render_str = Str((const char *)glGetString(GL_RENDERER));
  _version_str = Str((const char *)glGetString(GL_VERSION));
  DJV_DBG_PRN("OpenGL vendor string = " << _vendor_str);
  DJV_DBG_PRN("OpenGL renderer string = " << _render_str);
  DJV_DBG_PRN("OpenGL version string = " << _version_str);
  DJV_DBG_PRN("OpenGL extensions = " <<
    (const char *)glGetString(GL_EXTENSIONS));
  DJV_DBG_PRN("glu version = " <<
    (const char *)gluGetString(GLU_VERSION));
  DJV_DBG_PRN("glu extensions = " <<
    (const char *)gluGetString(GLU_EXTENSIONS));
}

Global::Global() throw (Error) :
  _device(new _Wgl_Device),
  _pipeline(PIPELINE(0))
{
  DJV_DBG_OBJ("Global::Global");
  
  _device->init();
  
  const Str & version = _device->version_str();
  if (version.size())
    if ('2' == version[0])
      _pipeline = PROGRAM;
  DJV_DBG_PRN("pipeline = " << _pipeline);
}

Global::~Global() {}

}
}

#endif
