/*------------------------------------------------------------------------------
 lib/djv_ui/choice.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include <djv_ui/choice.h>
#include <djv_ui/icon.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/label.h>
#include <djv_ui/style.h>
#include <FL/Fl.H>
#include <FL/Fl_Menu_Button.H>
#include <FL/fl_draw.H>

namespace djv_ui {
namespace choice {

namespace {

class _Menu_Button : public Fl_Menu_Button, public widget::Base
{
public:

  _Menu_Button();

  // Value.
  
  void set(int);
  int get() const { return _value; }
  Signal<int> sig;
  
  // List.
  
  void list(const List<Str> &);

  // Widget.
  
  int handle(int);
  void show();
  void draw();
  
  // Layout.
  
  void dirty();
  void geom(const Box2i &);
  const Box2i & geom() const { return widget::Base::geom(); }
  void resize(int x, int y, int w, int h) { geom(Box2i(x, y, w, h)); }
  int visible() const { return Fl_Menu_Button::visible(); }
  
  // Debugging.
  
  Str dbg_name() const { return "djv::ui::choice::_Menu_Button"; }

private:

  DJV_FL_WIDGET_CALLBACK(_Menu_Button, menu_callback) { set(value()); }

  void widget_update();

  int _value;
  List<mem::Buffer<char> > _list;
  Fl_Image * _icon;
};

_Menu_Button::_Menu_Button() :
  Fl_Menu_Button(0, 0, 0, 0),
  sig(this),
  _value(0),
  _icon(icon::get("tool_dec"))
{
  box(FL_THIN_UP_BOX);
  
  widget_update();
  
  callback(menu_callback, this);
}

void _Menu_Button::set(int in)
{
  if (in == _value) return;
  
  _value = in;
  
  redraw();
  sig.emit(_value);
}

Str pack(const Str & in)
{
  Str out = in;
  Str::size_type pos = 0;
  while ((pos = out.find('&', pos)) != Str::npos)
    out = out.replace(pos, 1, 0, ' ');
  return out;
}

void _Menu_Button::list(const List<Str> & in)
{
  _list.resize(in.size());
  for (uint i = 0; i < in.size(); ++i)
  {
    const Str & _in = pack(in[i]);
    mem::Buffer<char> & tmp = _list[i];
    tmp.size(_in.size() + 1);
    mem::copy(_in.c_str(), tmp(), _in.size() + 1);
  }
  
  _value = 0;
  
  widget_update();
  sig.emit(_value);
}

int _Menu_Button::handle(int in)
{
  switch (widget::Base::handle(in))
  {
    case -1: return 0;
    case 1: return 1;
  }
  return Fl_Menu_Button::handle(in);
}

void _Menu_Button::show()
{
  Fl_Menu_Button::show();
  widget::Base::show();
}

void _Menu_Button::draw()
{
  Fl_Menu_Button::draw_box(box(), color());
  fl_draw_box(box(), x(), y(), w(), h(), color());

  const int margin =  Fl::box_dx(box()) + 3;
  const Box2i geom(
    x() + margin,
    y() + margin,
    w() - margin * 2,
    h() - margin * 2);
  
  if (_value >= 0 && _value < int(_list.size()))
  {
    fl_font(textfont(), textsize());
    fl_color(FL_FOREGROUND_COLOR);
    fl_draw(_list[_value](), geom.x, geom.y + fl_height() - fl_descent());
  }
  
  if (_icon)
    _icon->draw(
      geom.x + geom.w - style::global()->size_tool_icon(),
      geom.y + geom.h / 2 - _icon->h() / 2);

  if (Fl::focus() == this) draw_focus();
}
  
void _Menu_Button::dirty()
{
  DJV_DBG_OBJ("_Menu_Button::dirty");
  
  widget::Base::dirty();

  textfont(text_font());
  textsize(style::global()->font_size());
  textcolor(text_color());
  
  fl_font(text_font(), style::global()->font_size());
  Vec2i size;
  for (uint i = 0; i < _list.size(); ++i)
    size = max(base::str_size(_list[i]()), size);
  
  size_hint(
    Vec2i(
      size.x + style::global()->size_tool_icon() +
        style::global()->margin_widget(),
      size.y
    ) +
    style::global()->margin_widget() * 2 + Fl::box_dx(box()) * 2);
}
  
void _Menu_Button::geom(const Box2i & in)
{
  Fl_Menu_Button::resize(in.x, in.y, in.w, in.h);
  widget::Base::geom(in);
}

void _Menu_Button::widget_update()
{
  callbacks(false);
  
  clear();
  const uint size = _list.size();
  mem::Buffer<Fl_Menu_Item> tmp(size + 1);
  mem::set(0, tmp(), (size + 1) * sizeof(Fl_Menu_Item));
  Fl_Menu_Item * p = tmp();
  for (uint i = 0; i < size; ++i)
  {
    p[i].text = _list[i]();
    p[i].callback_ = menu_callback;
    p[i].user_data_ = this;
  }
  copy(tmp());
  update();

  callbacks(true);
}

}

struct Widget::_Widget
{
  _Widget() :
    label(0)
  {}

  _Menu_Button * widget;
  widget::Widget * label;
  layout::Horizontal * layout;
};

Widget::Widget(const List<Str> & list, const Str & label) :
  sig(this),
  _p(new _Widget)
{
  DJV_DBG_OBJ("Widget::Widget");
  
  _p->widget = new _Menu_Button;
  if (! label.empty()) _p->label = new label::Widget(label);

  size(layout::FIXED, layout::FIXED);
  align(layout::LEFT, layout::CENTER);
  
  _p->layout = new Layout_H(this);
  _p->layout->margin(0);
  if (_p->label) _p->layout->add(_p->label);
  _p->layout->add(_p->widget);

  _p->widget->list(list);
  
  _p->widget->sig.set(this, widget_callback);
}

Widget::~Widget()
{}

void Widget::set(int in)
{
  _p->widget->set(in);
}

int Widget::get() const { return _p->widget->get(); }

void Widget::list(const List<Str> & in)
{
  if (in == _value) return;
  
  _value = in;
  
  _p->widget->list(_value);
}

const List<Str> & Widget::list() const { return _value; }

void Widget::label(widget::Widget * in)
{
  DJV_DBG_OBJ("Widget::label");
  
  if (_p->label) _p->layout->del(_p->label);
  _p->label = in;
  if (_p->label) _p->layout->insert(_p->label, 0);
  
  //update();
}

widget::Widget * Widget::label() const { return _p->label; }
layout::Item * Widget::label_align() { return _p->label; }

void Widget::text_font(const Fl_Font & in)
{
  widget::Widget::text_font(in);
  if (_p->label) _p->label->text_font(in);
}

const Fl_Font & Widget::text_font() const
{
  return widget::Widget::text_font();
}

void Widget::text_color(const Fl_Color & in)
{
  widget::Widget::text_color(in);
  if (_p->label) _p->label->text_color(in);
}

const Fl_Color & Widget::text_color() const
{
  return widget::Widget::text_color();
}

void Widget::widget_callback(int in) { sig.emit(in); }

}
}
