/*------------------------------------------------------------------------------
 lib/djv_ui/color_widget.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

#undef DJV_DBG
//#define DJV_DBG

#include <djv_ui/color.h>
#include <djv_ui/choice.h>
#include <djv_ui/layout_row.h>
#include <djv_ui/num_int.h>
#include <djv_ui/num_float.h>
#include <djv_ui/pixel.h>

namespace djv_ui {
namespace color {
using namespace djv_image::pixel;
using djv_image::Color;

struct Widget::_Widget
{
  num::Int_Slider * _int [djv_image::pixel::channels_max];
  num::Float_Slider * _float [djv_image::pixel::channels_max];
  choice::Widget * format;
  choice::Widget * type;
};

Widget::Widget() :
  sig(this),
  _value(RGB_U8),
  _float_min(0.0),
  _float_max(1.0),
  _p(new _Widget)
{
  // Widgets.

  for (int i = 0; i < channels_max; ++i)
  {
    _p->_int[i] = new num::Int_Slider;
    _p->_float[i] = new num::Float_Slider;
  }
  _p->format = new choice::Widget(djv_image::pixel::label_format());
  _p->type = new choice::Widget(djv_image::pixel::label_type());
  
  // Layout.

  Item::size(layout::STRETCH, layout::FIXED);
  
  List<layout::Item *> group;
  for (int i = 0; i < channels_max; ++i)
  {
    group += _p->_int[i];
    group += _p->_float[i];
  }
  layout::group(group);
  
  Layout_V * layout = new Layout_V(this);
  layout->margin(0);
  for (int i = 0; i < channels_max; ++i)
  {
    Layout_H * layout_h = new Layout_H(layout);
    layout_h->margin(0);
    layout_h->add(_p->_int[i]);
    layout_h->add(_p->_float[i]);
  }
  
  Layout_H * layout_h = new Layout_H(layout);
  layout_h->margin(0);
  layout_h->add_stretch();
  layout_h->add(_p->format);
  layout_h->add(_p->type);
  
  // Initialize.
  
  widget_update();
  value_update();
  
  // Callbacks.

  for (int i = 0; i < channels_max; ++i)
  {
    _p->_int[i]->sig.set(this, int_callback);
    _p->_float[i]->sig.set(this, float_callback);
  }
  _p->format->sig.set(this, format_callback);
  _p->type->sig.set(this, type_callback);
}

Widget::~Widget()
{}

void Widget::set(const Color & in)
{
  if (in == _value) return;
  
  DJV_DBG_OBJ("Widget::set");
  DJV_DBG_PRN("in = " << in);
  
  const Color prev = _value;
  _value = in;
  
  if (_value.pixel() != prev.pixel()) widget_update();
  value_update();
  sig.emit(_value);
}

const djv_image::Color & Widget::get() const { return _value; }

void Widget::float_range(double min, double max)
{
  if (min == _float_min && max == _float_max) return;
  
  _float_min = min;
  _float_max = max;
  
  set(_value);
  
  widget_update();
}

double Widget::float_min() const { return _float_min; }
double Widget::float_max() const { return _float_max; }

void Widget::int_callback(int)
{
  for (int i = 0; i < djv_image::pixel::channels(_value.pixel()); ++i)
    switch (djv_image::pixel::type(_value.pixel()))
    {
      case U8: _value.set_u8(_p->_int[i]->get(), i); break;
      case U10: _value.set_u10(_p->_int[i]->get(), i); break;
      case U16: _value.set_u16(_p->_int[i]->get(), i); break;
      case F16: _value.set_f16(_p->_float[i]->get(), i); break;
      case F32: _value.set_f32(_p->_float[i]->get(), i); break;
      default: break;
    }

  value_update();
  sig.emit(_value);
}

void Widget::float_callback(double) { int_callback(0); }

void Widget::format_callback(int in)
{
  Color tmp(djv_image::pixel::pixel(
    FORMAT(in),
    djv_image::pixel::type(_value.pixel())));
  djv_image::color::convert(_value, &tmp);
  
  set(tmp);
  widget_update();
}

void Widget::type_callback(int in)
{
  Color tmp(djv_image::pixel::pixel(
    format(_value.pixel()),
    TYPE(in)));
  djv_image::color::convert(_value, &tmp);
  
  set(tmp);
  widget_update();
}

void Widget::widget_update()
{
  DJV_DBG_OBJ("Widget::widget_update");
  DJV_DBG_PRN("value = " << _value);
  
  callbacks(false);
  
  const int channels = djv_image::pixel::channels(_value.pixel());
  
  const Str tooltip0 [] =
  {
    tooltip_luminance,
    tooltip_alpha
  };
  const Str tooltip1 [] =
  {
    tooltip_red,
    tooltip_green,
    tooltip_blue,
    tooltip_alpha
  };
  const Str * tooltip = 0;
  switch (djv_image::pixel::format(_value.pixel()))
  {
    case L:
    case LA: tooltip = tooltip0; break;
    case RGB:
    case RGBA: tooltip = tooltip1; break;
    default: break;
  }

  for (int i = 0; i < channels; ++i)
  {
    _p->_int[i]->range(0, djv_image::pixel::max(_value.pixel()));
    _p->_int[i]->tooltip(tooltip[i]);
    _p->_float[i]->range(_float_min, _float_max);
    _p->_float[i]->tooltip(tooltip[i]);
  }
 
  for (int i = 0; i < channels_max; ++i)
    switch (djv_image::pixel::type(_value.pixel()))
    {
      case U8:
      case U10:
      case U16:
        if (i < channels) _p->_int[i]->show();
        else _p->_int[i]->hide();
        _p->_float[i]->hide();
        break;
      case F16:
      case F32:
        _p->_int[i]->hide();
        if (i < channels) _p->_float[i]->show();
        else _p->_float[i]->hide();
        break;
      default: break;
    }
  
  _p->format->set(djv_image::pixel::format(_value.pixel()));
  _p->type->set(djv_image::pixel::type(_value.pixel()));

  update();

  callbacks(true);
}

void Widget::value_update()
{
  DJV_DBG_OBJ("Widget::value_update");
  
  callbacks(false);
  
  for (int i = 0; i < djv_image::pixel::channels(_value.pixel()); ++i)
    switch (djv_image::pixel::type(_value.pixel()))
    {
      case U8: _p->_int[i]->set(_value.get_u8(i)); break;
      case U10: _p->_int[i]->set(_value.get_u10(i)); break;
      case U16: _p->_int[i]->set(_value.get_u16(i)); break;
      case F16: _p->_float[i]->set(_value.get_f16(i)); break;
      case F32: _p->_float[i]->set(_value.get_f32(i)); break;
      default: break;
    }
  
  callbacks(true);
}

const Str
  tooltip_red = "Red channel.",
  tooltip_green = "Green channel.",
  tooltip_blue = "Blue channel.",
  tooltip_alpha = "Alpha channel.",
  tooltip_luminance = "Luminance channel.";

}
}
