/*------------------------------------------------------------------------------
 plugin/djv_jpeg/load.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io.h"

namespace djv_jpeg {
namespace io {
using libjpeg::jpeg_decompress_struct;

extern "C" {
DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_load() { return new Load; }
}

Load::Load() :
  _f(0),
  _jpeg_init(false)
{}

Load::~Load()
{
  _close();
}

plugin::Plugin * Load::copy() const { return new Load; }

const Str & Load::name() const { return djv_jpeg::name; }

const List<Str> & Load::extension_list() const
  { return djv_jpeg::extension_list; }

void Load::open(const File & in, Info * info, Tag * tag, djv_image::Time * time)
  throw (Error)
{
  DJV_DBG_OBJ("Load::open");
  DJV_DBG_PRN("in = " << in);

  _file = in;
  _open(_file.get(_file.seq.start()), info, tag);
  
  if (file::SEQ == _file.type)
    time->list = _file.seq.list;
}

void Load::open(const File & in, List<Info> * info, Tag * tag,
  djv_image::Time * time) throw (Error)
{
  djv_image::io::Load::open(in, info, tag, time);
}

namespace {

bool init(libjpeg::jpeg_decompress_struct * jpeg, Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::jpeg_create_decompress(jpeg);
  return true;
}

bool open(FILE * f, libjpeg::jpeg_decompress_struct * jpeg,
  Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::jpeg_stdio_src(jpeg, f);
  libjpeg::jpeg_save_markers(jpeg, JPEG_COM, 0xFFFF);
  if (! libjpeg::jpeg_read_header(jpeg, TRUE)) return false;
  if (! libjpeg::jpeg_start_decompress(jpeg)) return false;
  return true;
}

}

void Load::_open(const File & in, Info * info, Tag * tag) throw (Error)
{
  DJV_DBG_OBJ("Load::_open");
  DJV_DBG_PRN("in = " << in);

  _close();

  // Initialize libjpeg.
  
  _jpeg.err = libjpeg::jpeg_std_error(&_jpeg_error.pub);
  _jpeg_error.pub.error_exit = djv_jpeg_error;
  _jpeg_error.pub.emit_message = djv_jpeg_warn;
  _jpeg_error.msg[0] = 0;

  if (! io::init(&_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
  
  _jpeg_init = true;

  // Open.
  
  _f = ::fopen(in.get().c_str(), "rb");
  if (! _f)
    throw Error(name(), djv_image::io::error_open);

  if (! io::open(_f, &_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
  
  // Information.
  
  info->name = in;
  
  info->size = Vec2i(_jpeg.output_width, _jpeg.output_height);
  
  if (! pixel::pixel_(_jpeg.out_color_components, 8, false, &info->pixel))
    throw Error(name(), djv_image::io::error_unsupported);
  
  // Image tags.
  
  const libjpeg::jpeg_saved_marker_ptr marker = _jpeg.marker_list;
  if (marker)
    (*tag)[tag::label_tag()[tag::DESCRIPTION]] =
      Str((const char *)marker->data, marker->data_length);

  DJV_DBG_PRN("info = " << *info);
}

void Load::_close()
{
  if (_jpeg_init)
  {
    libjpeg::jpeg_destroy_decompress(&_jpeg);
    _jpeg_init = false;
  }
  if (_f) { ::fclose(_f); _f = 0; }
}

namespace {

bool scanline(libjpeg::jpeg_decompress_struct * jpeg, uint8_t * out,
  Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return 0;
  libjpeg::JSAMPROW p [] = { (libjpeg::JSAMPLE *)out };
  if (! libjpeg::jpeg_read_scanlines(jpeg, p, 1)) return false;
  return true;
}

bool end(libjpeg::jpeg_decompress_struct * jpeg, Jpeg_Error * error)
{
  if (::setjmp(error->jump)) return false;
  libjpeg::jpeg_finish_decompress(jpeg);
  return true;
}

}

const Image * Load::get(int frame, int, data::PROXY proxy) throw (Error)
{
  DJV_DBG_OBJ("Load::get");
  DJV_DBG_PRN("frame = " << frame);
  DJV_DBG_PRN("proxy = " << proxy);
  
  // Open.
  
  Info info;
  _open(_file.get(frame != -1 ? frame : _file.seq.start()), &info, &_image.tag);
  
  // Read.
  
  Data * data = proxy ? &_tmp : &_image;
  data->set(info);
  
  for (int y = 0; y < info.size.y; ++y)
    if (! scanline(&_jpeg, data->data(0, data->h() - 1 - y), &_jpeg_error))
      throw Error(name(), _jpeg_error.msg);
  
  if (! end(&_jpeg, &_jpeg_error))
    throw Error(name(), _jpeg_error.msg);
  
  if (proxy)
  {
    info.size = data::proxy_scale(info.size, proxy);
    info.proxy = proxy;
    _image.set(info);
    data::proxy_scale(_tmp, &_image, proxy);
  }

  DJV_DBG_PRN("image = " << _image);
  return &_image;
}

}
}

