/*------------------------------------------------------------------------------
 plugin/djv_tiff/save.cxx

 Copyright (c) 2004-2008 Darby Johnston. All rights reserved.
 Use is subject to license terms.
------------------------------------------------------------------------------*/

//#define DJV_DBG

#include "io.h"

namespace djv_tiff {
namespace io {

extern "C" {

DJV_PLUGIN_EXPORT plugin::Plugin * djv_image_save()
{
  tiff_init();

  return new Save;
}

}

Save::Save() :
  _f(0)
{}

Save::~Save()
{
  _close();
}

plugin::Plugin * Save::copy() const
{
  Save * out = new Save;
  out->_option = _option;
  return out;
}

const Str & Save::name() const { return djv_tiff::name; }

const List<Str> & Save::extension_list() const
  { return djv_tiff::extension_list; }

bool Save::option(const Str & in, Str * data)
{
try
{
  if (str::compare(in, option_list()[SAVE_COMPRESS])) *data >> _option.compress;
}
catch (Str) { return false; }
  return true;
}

Str Save::option(const Str & in) const
{
  Str out;
  if (str::compare(in, option_list()[SAVE_COMPRESS])) out << _option.compress;
  return out;
}

const List<Str> & Save::option_list() const { return label_save_option(); }

void Save::open(const File & in, const Info & info, const Tag &,
  const djv_image::Time & time) throw (Error)
{
  DJV_DBG_OBJ("Save::open");
  DJV_DBG_PRN("in = " << in);
  
  _file = in;
  if (time.list.size() > 1) _file.type = file::SEQ;
  
  _info = Info();
  _info.size = info.size;
  _info.mirror.y = true;
  pixel::TYPE type = pixel::type(info.pixel);
  switch (type)
  {
    case pixel::U10: type = pixel::U16; break;
    case pixel::F16: type = pixel::F32; break;
    default: break;
  }
  _info.pixel = pixel::pixel(pixel::format(info.pixel), type);
  DJV_DBG_PRN("info = " << _info);
  
  _image.set(_info);
}

void Save::_open(const File & in, const Tag & tag) throw (Error)
{
  DJV_DBG_OBJ("Save::_open");
  DJV_DBG_PRN("in = " << in);

  _close();
  
  // Open.
  
  _f = TIFFOpen(in.get().c_str(), "w");
  if (! _f)
    throw Error(name(), djv_image::io::error_open);
  
  // Header.
  
  uint16 photometric = 0, samples = 0, sample_depth = 0, sample_format = 0;
  uint16 extra_samples [] = { EXTRASAMPLE_ASSOCALPHA };
  uint16 extra_samples_size = 0;
  uint16 compress = 0;
  
  switch (pixel::format(_image.pixel()))
  {
    case pixel::L:
      photometric = PHOTOMETRIC_MINISBLACK;
      samples = 1;
      break;
    case pixel::LA:
      photometric = PHOTOMETRIC_MINISBLACK;
      samples = 2;
      extra_samples_size = 1;
      break;
    case pixel::RGB:
      photometric = PHOTOMETRIC_RGB;
      samples = 3;
      break;
    case pixel::RGBA:
      photometric = PHOTOMETRIC_RGB;
      samples = 4;
      extra_samples_size = 1;
      break;
    default: break;
  }
  
  switch (pixel::type(_image.pixel()))
  {
    case pixel::U8:
      sample_depth = 8;
      sample_format = SAMPLEFORMAT_UINT;
      break;
    case pixel::U16:
      sample_depth = 16;
      sample_format = SAMPLEFORMAT_UINT;
      break;
    case pixel::F16:
    case pixel::F32:
      sample_depth = 32;
      sample_format = SAMPLEFORMAT_IEEEFP;
      break;
    default: break;
  }
  
  switch (_option.compress)
  {
    case COMPRESS_NONE: compress = COMPRESSION_NONE; break;
    case COMPRESS_RLE: compress = COMPRESSION_PACKBITS; break;
    case COMPRESS_LZW: compress = COMPRESSION_LZW; break;
  }
  
  TIFFSetField(_f, TIFFTAG_IMAGEWIDTH, _image.w());
  TIFFSetField(_f, TIFFTAG_IMAGELENGTH, _image.h());
  TIFFSetField(_f, TIFFTAG_PHOTOMETRIC, photometric);
  TIFFSetField(_f, TIFFTAG_SAMPLESPERPIXEL, samples);
  TIFFSetField(_f, TIFFTAG_BITSPERSAMPLE, sample_depth);
  TIFFSetField(_f, TIFFTAG_SAMPLEFORMAT, sample_format);
  TIFFSetField(_f, TIFFTAG_EXTRASAMPLES, extra_samples_size, extra_samples);
  TIFFSetField(_f, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
  TIFFSetField(_f, TIFFTAG_COMPRESSION, compress);
  TIFFSetField(_f, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
  
  // Image tags.
  
  const List<Str> & label_tag = tag::label_tag();
  Str _tag = tag[label_tag[tag::CREATOR]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_ARTIST, _tag.c_str());
  _tag = tag[label_tag[tag::COPYRIGHT]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_COPYRIGHT, _tag.c_str());
  _tag = tag[label_tag[tag::TIME]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_DATETIME, _tag.c_str());
  _tag = tag[label_tag[tag::DESCRIPTION]];
  if (_tag.size()) TIFFSetField(_f, TIFFTAG_IMAGEDESCRIPTION, _tag.c_str());
}

void Save::_close()
{
  if (_f) { TIFFClose(_f); _f = 0; }
}

void Save::set(const Image & in, int frame) throw (Error)
{
  DJV_DBG_OBJ("Save::set");
  DJV_DBG_PRN("in = " << in);

  // Open.
  
  _open(_file.get(frame), in.tag);
  
  // Write.
  
  const Data * p = &in;
  if (in.info() != _info)
  {
    DJV_DBG_PRN("convert = " << _image);
    _image.zero();
    gl_data::copy(in, &_image);
    p = &_image;
  }
  
  const int h = p->h();
  for (int y = 0; y < h; ++y)
    if (TIFFWriteScanline(_f, (tdata_t *)p->data(0, y), y) == -1)
      throw Error(name(), djv_image::io::error_write);

  _close();
}

const List<Str> & label_save_option()
{
  static const List<Str> data = List<Str>() <<
    "Compress" <<
    "Channels";
  return data;
}

}
}

