#define DJV_DBG

#include "widget.h"
#include <djv_ui/style.h>
#include <FL/Fl_Scrollbar.h>
#include <FL/fl_draw.H>

namespace misc {

namespace {

const int frame_width = 10;

class _Widget : public Fl_Widget, public djv_ui::Callback_Base
{
public:

  _Widget();
  
  void set(Project *);
  Project * get() const;
  djv_ui::Signal<Project *> sig;
  
  void view(const Vec2i &);
  
  void draw();
  
private:

  Project * _project;
  Vec2i _view;
};

_Widget::_Widget() :
  Fl_Widget(0, 0, 0, 0),
  sig(this),
  _project(0)
{}

void _Widget::set(Project * in)
{
  _project = in;
}

Project * _Widget::get() const { return _project; }

void _Widget::view(const Vec2i & in)
{
  _view = in;
}

namespace {

void seconds_to_time(
  int in, 
  int * hour,
  int * minute,
  int * second
) {
  while (in > 60 * 60)
  {
    (*hour)++;
    in -= 60 * 60;
  }
  while (in > 60)
  {
    (*hour)++;
    in -= 60;
  }
  *second = in;
}

}

void _Widget::draw()
{
  DJV_DBG_OBJ("_Widget::draw");
  DJV_DBG_PRN("view = " << _view);
  
  const Box2i geom(x(), y(), w(), h());
  DJV_DBG_PRN("size = " << geom.size);
  
  const int margin = djv_ui::style::global()->margin();
  
  fl_rectf(geom.x, geom.y, geom.w, geom.h, FL_BACKGROUND2_COLOR);
  
  if (! _project) return;
  
  {
    fl_color(fl_lighter(FL_BACKGROUND2_COLOR));
    
    const int start = _view.x / frame_width;
    int x = _view.x - (start * frame_width);
    const int size = (geom.w - x) / frame_width + 1;
    DJV_DBG_PRN("frames = " << start << " / " << size);
    for (int i = 0; i < size; ++i, x += frame_width)
      fl_line(x, geom.y, x, geom.y + geom.h);
  }
  
  {
    fl_color(FL_WHITE);
    fl_font(
      djv_ui::style::global()->font(),
      djv_ui::style::global()->font_size()
      );
    
    const double w =
      frame_width * djv_image::time::speed_to_float(_project->time.speed);
    const int start = int(_view.x / w);
    int x = _view.x - int(start * w);
    const int size = int((geom.w - x) / w) + 1;
    DJV_DBG_PRN("seconds = " << start << " / " << size);
    for (int i = 0; i < size; ++i, x += int(w))
    {
      fl_line(x, geom.y, x, geom.y + geom.h);
      
      int hour = 0, minute = 0, second = 0;
      seconds_to_time(start + i, &hour, &minute, &second);
      const Str str = djv_image::time::timecode_to_str(
        djv_image::time::time_to_timecode(hour, minute, second, 0));
      
      int _x = x + margin;
      int _y = geom.y + fl_height() - fl_descent();
      fl_color(FL_WHITE);
      fl_draw(str.c_str(), _x, _y);
    }
  }
}

}

struct Timeline_Widget::P
{
  _Widget * widget;
  Fl_Scrollbar * scrollbar_h, * scrollbar_v;
};

Timeline_Widget::Timeline_Widget() :
  sig(this),
  _project(0),
  _p(new P)
{
  _p->widget = new _Widget;
  Fl_Group::add(_p->widget);
  _p->scrollbar_h = new Fl_Scrollbar(0, 0, 0, 0);
  _p->scrollbar_h->type(FL_HORIZONTAL);
  Fl_Group::add(_p->scrollbar_h);
  _p->scrollbar_v = new Fl_Scrollbar(0, 0, 0, 0);
  Fl_Group::add(_p->scrollbar_v);
  
  widget_update();
  
  _p->scrollbar_h->callback(scrollbar_callback, this);
  _p->scrollbar_v->callback(scrollbar_callback, this);
}

Timeline_Widget::~Timeline_Widget()
{}

void Timeline_Widget::set(Project * in)
{
  _project = in;
  
  _p->widget->set(in);
  _p->widget->redraw();

  widget_update();
}

Project * Timeline_Widget::get() const { return _project; }

void Timeline_Widget::dirty()
{
  djv_ui::widget::Widget::dirty();

  size_hint(
    Vec2i(600, 300) +
    Vec2i(16, 16)
    );
}

void Timeline_Widget::geom(const Box2i & in)
{
  djv_ui::widget::Widget::geom(in);
  
  _p->widget->resize(in.x, in.y, in.w - 16, in.h - 16);
  _p->scrollbar_h->resize(in.x, in.y + in.h - 16, in.w - 16, 16);
  _p->scrollbar_v->resize(in.x + in.w - 16, in.y, 16, in.h - 16);
  
  widget_update();
}

const Box2i & Timeline_Widget::geom() const
{
  return djv_ui::widget::Widget::geom();
}

void Timeline_Widget::scrollbar_callback()
{
  widget_update();
}

void Timeline_Widget::widget_update()
{
  DJV_DBG_OBJ("Timeline_Widget::widget_update");
  
  callbacks(false);
  
  Vec2i size;
  if (_project)
  {
    size.x = _project->time.size * frame_width;
    size.y = 600;
  }
  DJV_DBG_PRN("size = " << size);
  
  Vec2i position(_p->scrollbar_h->value(), _p->scrollbar_v->value());
  DJV_DBG_PRN("position = " << position);
  
  _p->widget->view(position);
  _p->widget->redraw();
  _p->scrollbar_h->value(position.x, _p->widget->w(), 0, size.x);
  _p->scrollbar_v->value(position.y, _p->widget->h(), 0, size.y);
  
  callbacks(true);
}

}
